#  Optimized algorithm to search for transits of small extrasolar planets
#                                                                            /
#       ,        AUTHORS                                                   O/
#    \  :  /     Michael Hippke (1) [michael@hippke.org]                /\/|
# `. __/ \__ .'  Rene' Heller (2) [heller@mps.mpg.de]                      |
# _ _\     /_ _  _________________________________________________________/ \_
#    /_   _\
#  .'  \ /  `.   (1) Sonneberg Observatory, Sternwartestr. 32, Sonneberg
#    /  :  \     (2) Max Planck Institute for Solar System Research,
#       '            Justus-von-Liebig-Weg 3, 37077 G\"ottingen, Germany

from __future__ import division, print_function
from .main import transitleastsquares
from .helpers import cleaned_array, resample, transit_mask
from .stats import FAP
from .catalog import catalog_info
from .core import fold
from .template_generator.transit_template_generator import TransitTemplateGenerator
from .template_generator.default_transit_template_generator import DefaultTransitTemplateGenerator
from .results import transitleastsquaresresults
from . import version



from __future__ import division, print_function
import numpy
from foldedleastsquares import cleaned_array

if __name__ == "__main__":
    print("Starting test: cleaned_array...", end="")

    dirty_array = numpy.ones(10, dtype=object)
    time_array = numpy.linspace(1, 10, 10)
    dy_array = numpy.ones(10, dtype=object)
    dirty_array[1] = None
    dirty_array[2] = numpy.inf
    dirty_array[3] = -numpy.inf
    dirty_array[4] = numpy.nan
    dirty_array[5] = -99
    time_array[8] = numpy.nan
    dy_array[9] = numpy.inf

    t, y, dy = cleaned_array(time_array, dirty_array, dy_array)
    numpy.testing.assert_equal(t, [1, 7, 8])
    numpy.testing.assert_equal(y, [1, 1, 1])
    numpy.testing.assert_equal(dy, [1, 1, 1])
    print("passed")



from __future__ import division, print_function
import numba
import numpy
from numpy import arccos, degrees
from .interpolation import interp1d


def resample(time, flux, factor):
    # New method without scipy
    time_grid = int(len(flux) / factor)
    time_resampled = numpy.linspace(min(time), max(time), time_grid)
    f = interp1d(time_resampled, time)
    flux_resampled = f(flux)

    return time_resampled, flux_resampled


def cleaned_array(t, y, dy=None):
    """Takes numpy arrays with masks and non-float values.
    Returns unmasked cleaned arrays."""

    def isvalid(value):
        valid = False
        if value is not None:
            if not numpy.isnan(value):
                if value > 0 and value < numpy.inf:
                    valid = True
        return valid

    # Start with empty Python lists and convert to numpy arrays later (reason: speed)
    clean_t = []
    clean_y = []
    if dy is not None:
        clean_dy = []

    # Cleaning numpy arrays with both NaN and None values is not trivial, as the usual
    # mask/delete filters do not accept their simultanous ocurrence without warnings.
    # Instead, we iterate over the array once; this is not Pythonic but works reliably.
    for i in range(len(y)):

        # Case: t, y, dy
        if dy is not None:
            if isvalid(y[i]) and isvalid(t[i]) and isvalid(dy[i]):
                clean_y.append(y[i])
                clean_t.append(t[i])
                clean_dy.append(dy[i])

        # Case: only t, y
        else:
            if isvalid(y[i]) and isvalid(t[i]):
                clean_y.append(y[i])
                clean_t.append(t[i])

    clean_t = numpy.array(clean_t, dtype=float)
    clean_y = numpy.array(clean_y, dtype=float)

    if dy is None:
        return clean_t, clean_y
    else:
        clean_dy = numpy.array(clean_dy, dtype=float)
        return clean_t, clean_y, clean_dy


def transit_mask(t, period, duration, T0):
    # Works with numba, but is not faster
    mask = numpy.abs((t - T0 + 0.5 * period) % period - 0.5 * period) < 0.5 * duration
    return mask


def running_mean(data, width_signal):
    """Returns the running mean in a given window"""
    cumsum = numpy.cumsum(numpy.insert(data, 0, 0))
    return (cumsum[width_signal:] - cumsum[:-width_signal]) / float(width_signal)


def running_mean_equal_length(data, width_signal):
    """Returns the running mean in a given window"""
    cumsum = numpy.cumsum(numpy.insert(data, 0, 0))
    med = (cumsum[width_signal:] - cumsum[:-width_signal]) / float(width_signal)

    # Append the first/last value at the beginning/end to match the length of
    # data and returned median
    first_values = med[0]
    last_values = med[-1]
    missing_values = len(data) - len(med)
    values_front = int(missing_values * 0.5)
    values_end = missing_values - values_front
    med = numpy.append(numpy.full(values_front, first_values), med)
    med = numpy.append(med, numpy.full(values_end, last_values))
    return med


def running_median(data, kernel):
    """Returns sliding median of width 'kernel' and same length as data """
    idx = numpy.arange(kernel) + numpy.arange(len(data) - kernel + 1)[:, None]
    med = numpy.median(data[idx], axis=1)

    # Append the first/last value at the beginning/end to match the length of
    # data and returned median
    first_values = med[0]
    last_values = med[-1]
    missing_values = len(data) - len(med)
    values_front = int(missing_values * 0.5)
    values_end = missing_values - values_front
    med = numpy.append(numpy.full(values_front, first_values), med)
    med = numpy.append(med, numpy.full(values_end, last_values))
    return med


def impact_to_inclination(b, semimajor_axis):
    """Converts planet impact parameter b = [0..1.x] to inclination [deg]"""
    return degrees(arccos(b / semimajor_axis))



from __future__ import division, print_function
import numba
from . import tls_constants as tls_constants
import numpy
from numpy import pi, sqrt, arccos, degrees, floor, ceil
import warnings


@numba.jit(fastmath=True, parallel=False, nopython=True)
def T14(
    R_s, M_s, P, upper_limit=tls_constants.FRACTIONAL_TRANSIT_DURATION_MAX, small=False
):
    """Input:  Stellar radius and mass; planetary period
               Units: Solar radius and mass; days
       Output: Maximum planetary transit duration T_14max
               Unit: Fraction of period P"""

    P = P * tls_constants.SECONDS_PER_DAY
    R_s = tls_constants.R_sun * R_s
    M_s = tls_constants.M_sun * M_s

    if small:  # small planet assumption
        T14max = R_s * ((4 * P) / (pi * tls_constants.G * M_s)) ** (1 / 3)
    else:  # planet size 2 R_jup
        T14max = (R_s + 2 * tls_constants.R_jup) * (
            (4 * P) / (pi * tls_constants.G * M_s)
        ) ** (1 / 3)

    result = T14max / P
    if result > upper_limit:
        result = upper_limit
    return result



from __future__ import division, print_function
import numpy
import numba
from . import tls_constants
from .helpers import running_mean


@numba.jit(fastmath=True, parallel=False, nopython=True)
def fold(time, period, T0):
    """Normal phase folding"""
    return (time - T0) / period - numpy.floor((time - T0) / period)


@numba.jit(fastmath=True, parallel=False, nopython=True)
def foldfast(time, period):
    """Fast phase folding with T0=0 hardcoded"""
    return time / period - numpy.floor(time / period)


@numba.jit(fastmath=True, parallel=False, nopython=True)
def edge_effect_correction(flux, patched_data, dy, inverse_squared_patched_dy):
    regular = numpy.sum(((1 - flux) ** 2) * 1 / dy ** 2)
    patched = numpy.sum(((1 - patched_data) ** 2) * inverse_squared_patched_dy)
    return patched - regular


@numba.jit(fastmath=True, parallel=False, nopython=True)
def lowest_residuals_in_this_duration(
    mean,
    transit_depth_min,
    patched_data_arr,
    duration,
    signal,
    inverse_squared_patched_dy_arr,
    overshoot,
    ootr,
    summed_edge_effect_correction,
    chosen_transit_row,
    datapoints,
    T0_fit_margin,
):

    # if nothing is fit, we fit a straight line: signal=1. Then, at dy=1,
    # the squared sum of residuals equals the number of datapoints
    summed_residual_in_rows = datapoints
    best_row = 0
    best_depth = 0
    mean = 1 - mean
    xth_point = 1  # How many cadences the template shifts forward in each step
    if T0_fit_margin > 0 and duration > T0_fit_margin:
        xth_point = int(duration * T0_fit_margin)
        if xth_point < 1:
            xth_point = 1

    for i in range(len(mean)):
        if (mean[i] > transit_depth_min) and (i % xth_point == 0):
            data = patched_data_arr[i: i + duration]
            dy = inverse_squared_patched_dy_arr[i: i + duration]
            target_depth = mean[i] * overshoot
            reverse_scale = 1 * target_depth / tls_constants.SIGNAL_DEPTH  # speed: one division now, many mults later

            # Scale model and calculate residuals
            current_stat = numpy.sum((data - (1 - ((1 - signal) * reverse_scale))) ** 2 * dy) + ootr[i] - summed_edge_effect_correction
            if current_stat < summed_residual_in_rows:
                summed_residual_in_rows = current_stat
                best_row = chosen_transit_row
                best_depth = 1 - target_depth

    return summed_residual_in_rows, best_row, best_depth


@numba.jit(fastmath=True, parallel=False, nopython=True)
def out_of_transit_residuals(data, width_signal, dy):
    chi2 = numpy.zeros(len(data) - width_signal + 1)
    fullsum = numpy.sum(((1 - data) ** 2) * dy)
    window = numpy.sum(((1 - data[:width_signal]) ** 2) * dy[:width_signal])
    chi2[0] = fullsum - window
    for i in range(1, len(data) - width_signal + 1):
        becomes_visible = i - 1
        becomes_invisible = i - 1 + width_signal
        add_visible_left = (1 - data[becomes_visible]) ** 2 * dy[becomes_visible]
        remove_invisible_right = (1 - data[becomes_invisible]) ** 2 * dy[
            becomes_invisible
        ]
        chi2[i] = chi2[i - 1] + add_visible_left - remove_invisible_right
    return chi2


def search_period(
    transit_template_generator,
    period,
    t,
    y,
    dy,
    transit_depth_min,
    R_star_min,
    R_star_max,
    M_star_min,
    M_star_max,
    lc_arr,
    lc_cache_overview,
    T0_fit_margin,
):
    """Core routine to search the flux data set 'injected' over all 'periods'"""

    # duration (in samples) of widest transit in lc_cache (axis 0: rows; axis 1: columns)
    durations = numpy.unique(lc_cache_overview["width_in_samples"])
    maxwidth_in_samples = int(max(durations))
    if maxwidth_in_samples % 2 != 0:
        maxwidth_in_samples = maxwidth_in_samples + 1

    # Phase fold
    phases = foldfast(t, period)
    sort_index = numpy.argsort(phases, kind="mergesort")  # 8% faster than Quicksort
    phases = phases[sort_index]
    flux = y[sort_index]
    dy = dy[sort_index]

    # faster to multiply than divide
    patched_dy = numpy.append(dy, dy[:maxwidth_in_samples])
    inverse_squared_patched_dy = 1 / patched_dy ** 2

    # Due to phase folding, the signal could start near the end of the data
    # and continue at the beginning. To avoid (slow) rolling,
    # we patch the beginning again to the end of the data
    patched_data = numpy.append(flux, flux[:maxwidth_in_samples])

    this_edge_effect_correction = edge_effect_correction(
        flux, patched_data, dy, inverse_squared_patched_dy
    )

    # Set "best of" counters to max, in order to find smaller residuals
    smallest_residuals_in_period = float("inf")
    summed_residual_in_rows = float("inf")

    # Make unique to avoid duplicates in dense grids
    duration_max = transit_template_generator.max_duration(period, R_star_max, M_star_max)
    duration_min = transit_template_generator.min_duration(period, R_star_min, M_star_min)

    # Fractional transit duration can be longer than this.
    # Example: Data length 11 days, 2 transits at 0.5 days and 10.5 days
    length = max(t) - min(t)
    no_of_transits_naive = length / period
    no_of_transits_worst = no_of_transits_naive + 1
    correction_factor = no_of_transits_worst / no_of_transits_naive

    duration_min_in_samples = int(numpy.floor(duration_min * len(y)))
    duration_max_in_samples = int(numpy.ceil(duration_max * len(y) * correction_factor))
    durations = durations[durations >= duration_min_in_samples]
    durations = durations[durations <= duration_max_in_samples]

    skipped_all = True
    best_row = 0  # shortest and shallowest transit
    best_depth = 0

    for duration in durations:
        chosen_transit_row = 0
        while lc_cache_overview["width_in_samples"][chosen_transit_row] != duration:
            chosen_transit_row += 1
        this_residual, this_row, this_depth = lowest_residuals_in_this_duration(
            mean=running_mean(patched_data, duration),
            transit_depth_min=transit_depth_min,
            patched_data_arr=patched_data,
            duration=duration,
            signal=lc_arr[chosen_transit_row],
            inverse_squared_patched_dy_arr=inverse_squared_patched_dy,
            overshoot=lc_cache_overview["overshoot"][chosen_transit_row],
            ootr=out_of_transit_residuals(
                patched_data, duration, inverse_squared_patched_dy
            ),
            summed_edge_effect_correction=this_edge_effect_correction,
            chosen_transit_row=chosen_transit_row,
            datapoints=len(flux),
            T0_fit_margin=T0_fit_margin,
        )

        if this_residual < summed_residual_in_rows:
            summed_residual_in_rows = this_residual
            best_row = chosen_transit_row
            best_depth = this_depth

    return [period, summed_residual_in_rows, best_row, best_depth]



from __future__ import division, print_function
import numpy
import scipy
import scipy.signal
from foldedleastsquares import transitleastsquares
from foldedleastsquares.template_generator.default_transit_template_generator import DefaultTransitTemplateGenerator


def loadfile(filename):
    data = numpy.genfromtxt(filename, delimiter=",", dtype="f8, f8", names=["t", "y"])
    return data["t"], data["y"]

class TestTransitTemplateGenerator(DefaultTransitTemplateGenerator):
    def __init__(self):
        super().__init__()

if __name__ == "__main__":
    print("Starting test: transit shapes...", end="")

    # Testing transit shapes
    t, y = loadfile("EPIC206154641.csv")
    trend = scipy.signal.medfilt(y, 25)
    y_filt = y / trend

    # box
    model_box = transitleastsquares(t, y_filt)
    results_box = model_box.power(transit_template="box", period_grid=numpy.arange(1, 5, 0.01))
    numpy.testing.assert_equal(len(model_box.period_grid), 400)

    results_box = model_box.power(transit_template="box")
    numpy.testing.assert_almost_equal(
        results_box.duration, 0.06207881501022775, decimal=5)
    numpy.testing.assert_almost_equal(results_box.rp_rs, 0.08836981203437415, decimal=5)
    print("Test passed: Box-shaped")

    # grazing
    model_grazing = transitleastsquares(t, y_filt)
    results_grazing = model_grazing.power(transit_template="grazing")

    numpy.testing.assert_almost_equal(
        results_grazing.duration, 0.09079026695245815, decimal=5
    )
    numpy.testing.assert_almost_equal(
        min(results_grazing.chi2red), 0.06759475703796078, decimal=5)
    print("Test passed: Grazing-shaped")

    # comet
    model_comet = transitleastsquares(t, y_filt)
    results_comet = model_comet.power(transit_template="tailed")
    numpy.testing.assert_almost_equal(
        results_comet.duration, 0.23745146741412124, decimal=5
    )
    numpy.testing.assert_almost_equal(
        min(results_comet.chi2red), 0.0980794344892094, decimal=5)
    print("Test passed: Comet-shaped")

    model_custom = transitleastsquares(t, y_filt)
    try:
        results_custom = model_custom.power(transit_template="custom",
                                            transit_template_generator="wrongTransitTemplateGenerator")
        assert False
    except ValueError as e:
        if e.args[0] == "The custom transit_template_generator does not implement TransitTemplateGenerator.":
            print("Test passed: Wrong custom transit template generator.")
        else:
            assert False

    # custom
    model_custom = transitleastsquares(t, y_filt)
    results_custom = model_custom.power(transit_template="custom",
                                        transit_template_generator=TestTransitTemplateGenerator())
    numpy.testing.assert_almost_equal(
        results_custom.duration, 0.06828669651125058, decimal=5
    )
    numpy.testing.assert_almost_equal(
        min(results_custom.chi2red), 0.09977336183179186, decimal=5)
    print("Test passed: Custom-shaped")

    print("All tests passed")



from __future__ import division, print_function
import numpy
from foldedleastsquares import resample

if __name__ == "__main__":
    print("Starting test resample...", end="")

    # Resampling
    testtime = numpy.linspace(0, 1, 1000)
    testflux = numpy.linspace(0.99, 1.01, 1000)
    a, b = resample(time=testtime, flux=testflux, factor=100)
    expected_result_a = (
        0.0,
        0.11111111,
        0.22222222,
        0.33333333,
        0.44444444,
        0.55555556,
        0.66666667,
        0.77777778,
        0.88888889,
        1.0,
    )
    numpy.testing.assert_almost_equal(a, expected_result_a)
    expected_result_b = (
        0.99,
        0.99222222,
        0.99444444,
        0.99666667,
        0.99888889,
        1.00111111,
        1.00333333,
        1.00555556,
        1.00777778,
        1.01,
    )
    numpy.testing.assert_almost_equal(b, expected_result_b)
    numpy.testing.assert_equal(len(a), 10)
    numpy.testing.assert_almost_equal(min(a), 0)
    numpy.testing.assert_almost_equal(max(a), 1)
    numpy.testing.assert_almost_equal(numpy.mean(a), 0.5)
    numpy.testing.assert_equal(len(b), 10)
    numpy.testing.assert_almost_equal(min(b), 0.99)
    numpy.testing.assert_almost_equal(max(b), 1.01)
    numpy.testing.assert_almost_equal(numpy.mean(b), 1)
    print("passed")



from __future__ import division, print_function
import numpy
from foldedleastsquares import catalog_info
import sys

    
    

if __name__ == "__main__":
    if sys.version_info >= (3, 0):
        print("Starting test: catalog_data...")
        (a, b), mass, mass_min, mass_max, radius, radius_min, radius_max = catalog_info(
            EPIC_ID=204341806
        )
        numpy.testing.assert_almost_equal((a, b), (1.1605, -0.1704))
        numpy.testing.assert_almost_equal(mass, numpy.nan)
        numpy.testing.assert_almost_equal(mass_min, numpy.nan)
        numpy.testing.assert_almost_equal(mass_max, numpy.nan)
        numpy.testing.assert_almost_equal(radius, numpy.nan)
        numpy.testing.assert_almost_equal(radius_min, numpy.nan)
        numpy.testing.assert_almost_equal(radius_max, numpy.nan)

        (a, b), mass, mass_min, mass_max, radius, radius_min, radius_max = catalog_info(
            EPIC_ID=204099713
        )
        numpy.testing.assert_almost_equal((a, b), (0.4804, 0.1867))
        numpy.testing.assert_almost_equal(mass, 1.046)
        numpy.testing.assert_almost_equal(mass_min, 0.898)
        numpy.testing.assert_almost_equal(mass_max, 0.642)
        numpy.testing.assert_almost_equal(radius, 1.261)
        numpy.testing.assert_almost_equal(radius_min, 1.044)
        numpy.testing.assert_almost_equal(radius_max, 0.925)
        print("Test passed: EPIC catalog pull from Vizier using astroquery")

        (a, b), mass, mass_min, mass_max, radius, radius_min, radius_max = catalog_info(
            KIC_ID=757076
        )
        numpy.testing.assert_almost_equal((a, b), (0.5819, 0.137))
        numpy.testing.assert_almost_equal(mass, 1.357)
        numpy.testing.assert_almost_equal(mass_max, 0.204)
        numpy.testing.assert_almost_equal(mass_min, 0.475)
        numpy.testing.assert_almost_equal(radius, 3.128)
        numpy.testing.assert_almost_equal(radius_max, 0.987)
        numpy.testing.assert_almost_equal(radius_min, 2.304)

        (a, b), mass, mass_min, mass_max, radius, radius_min, radius_max = catalog_info(
            KIC_ID=12557548
        )
        numpy.testing.assert_almost_equal((a, b), (0.701, 0.0462))
        numpy.testing.assert_almost_equal(mass, 0.6666, decimal=3)
        numpy.testing.assert_almost_equal(mass_max, 0.067, decimal=3)
        numpy.testing.assert_almost_equal(mass_min, 0.055, decimal=3)
        numpy.testing.assert_almost_equal(radius, 0.660, decimal=3)
        numpy.testing.assert_almost_equal(radius_max, 0.054, decimal=3)
        numpy.testing.assert_almost_equal(radius_min, 0.054, decimal=3)
        print("Test passed: KIC catalog pull from Vizier using astroquery")

        (a, b), mass, mass_min, mass_max, radius, radius_min, radius_max = catalog_info(
            TIC_ID=279741377
        )
        numpy.testing.assert_equal((a, b), (0.354, 0.2321))
        numpy.testing.assert_equal(mass, 1.12)
        numpy.testing.assert_equal(mass_min, 0.148162)
        numpy.testing.assert_equal(mass_max, 0.148162)
        numpy.testing.assert_equal(radius, 3.08104)
        numpy.testing.assert_equal(radius_min, 0.167789)
        numpy.testing.assert_equal(radius_max, 0.167789)
        print("Test passed: TESS Input Catalog (TIC) pull from Vizier using astroquery")

        print("All tests passed")
    else:
        print('Skipping test, not working on Python 2')




from __future__ import division, print_function
import numpy
import batman
from foldedleastsquares import transitleastsquares


if __name__ == "__main__":
    print("Starting test: transit_depth_min...", end="")
    # Test for transit_depth_min=1000*10**-6, where no transit is found

    # Create test data
    start = 48
    days = 365.25 * 3
    samples_per_day = 12  # 48
    samples = int(days * samples_per_day)  # 48
    t = numpy.linspace(start, start + days, samples)

    # Use batman to create transits
    ma = batman.TransitParams()
    ma.t0 = (
        start + 20
    )  # time of inferior conjunction; first transit is X days after start
    ma.per = 365.25  # orbital period
    ma.rp = 6371 / 696342  # 6371 planet radius (in units of stellar radii)
    ma.a = 217  # semi-major axis (in units of stellar radii)
    ma.inc = 90  # orbital inclination (in degrees)
    ma.ecc = 0  # eccentricity
    ma.w = 90  # longitude of periastron (in degrees)
    ma.u = [0.5]  # limb darkening coefficients
    ma.limb_dark = "linear"  # limb darkening model
    m = batman.TransitModel(ma, t)  # initializes model
    original_flux = m.light_curve(ma)  # calculates light curve

    # Create noise and merge with flux
    ppm = 5
    stdev = 10 ** -6 * ppm
    noise = numpy.random.normal(0, stdev, int(samples))
    y = original_flux + noise
    y[1] = numpy.nan
    model = transitleastsquares(t, y)
    results = model.power(
        transit_depth_min=1000 * 10 ** -6,
        period_min=360,
        period_max=370,
        oversampling_factor=5,
        duration_grid_step=1.02,
        T0_fit_margin=0.1,
    )

    numpy.testing.assert_equal(results.transit_times, numpy.nan)
    numpy.testing.assert_equal(results.period, numpy.nan)
    numpy.testing.assert_equal(results.depth, 1)
    numpy.testing.assert_equal(results.duration, numpy.nan)
    numpy.testing.assert_equal(results.snr, numpy.nan)
    numpy.testing.assert_equal(results.snr_pink_per_transit, numpy.nan)
    numpy.testing.assert_equal(results.odd_even_mismatch, numpy.nan)
    numpy.testing.assert_equal(results.SDE, 0)
    numpy.testing.assert_equal(results.SDE_raw, 0)
    numpy.testing.assert_equal(results.in_transit_count, numpy.nan)
    numpy.testing.assert_equal(results.after_transit_count, numpy.nan)
    numpy.testing.assert_equal(results.before_transit_count, numpy.nan)
    numpy.testing.assert_almost_equal(results.chi2_min, 13148.0)
    numpy.testing.assert_almost_equal(results.chi2red_min, 1.0003043213633598)
    numpy.testing.assert_equal(len(results.periods), 278)
    numpy.testing.assert_almost_equal(max(results.periods), 369.9831654894093)
    numpy.testing.assert_almost_equal(min(results.periods), 360.0118189140635)
    numpy.testing.assert_almost_equal(max(results.power), 0)
    numpy.testing.assert_almost_equal(min(results.periods), 360.0118189140635)
    numpy.testing.assert_almost_equal(min(results.power), 0)
    numpy.testing.assert_almost_equal(max(results.chi2), 13148.0)
    numpy.testing.assert_almost_equal(max(results.chi2red), 1.0003043213633598)
    print("Test passed: transit_depth_min=1000*10**-6, where no transit is fit")



from __future__ import division, print_function
import numpy

from foldedleastsquares import DefaultTransitTemplateGenerator

if __name__ == "__main__":
    print("Starting test: period_grid...", end="")
    default_transit_template_generator = DefaultTransitTemplateGenerator()
    periods = default_transit_template_generator.period_grid(R_star=1, M_star=1, time_span=0.1)
    numpy.testing.assert_almost_equal(max(periods), 2.4999999999999987)
    numpy.testing.assert_almost_equal(min(periods), 0.6002621413799498)
    numpy.testing.assert_equal(len(periods), 268)

    periods = default_transit_template_generator.period_grid(R_star=1, M_star=1, time_span=20)
    numpy.testing.assert_almost_equal(max(periods), 10)
    numpy.testing.assert_almost_equal(min(periods), 0.6015575922909607)
    numpy.testing.assert_equal(len(periods), 1716)

    periods = default_transit_template_generator.period_grid(
        R_star=5,  # R_sun
        M_star=1,  # M_sun
        time_span=20,  # days
        period_min=0,
        period_max=999,
        oversampling_factor=3,
    )
    numpy.testing.assert_almost_equal(max(periods), 10)
    numpy.testing.assert_almost_equal(min(periods), 0.6015575922909607)
    numpy.testing.assert_equal(len(periods), 1716)

    periods = default_transit_template_generator.period_grid(
        R_star=1,  # R_sun
        M_star=1,  # M_sun
        time_span=20,  # days
        period_min=0,
        period_max=999,
        oversampling_factor=3,
    )
    numpy.testing.assert_almost_equal(max(periods), 10)
    numpy.testing.assert_almost_equal(min(periods), 0.60155759)
    numpy.testing.assert_equal(len(periods), 1716)

    periods = default_transit_template_generator.period_grid(
        R_star=0.1,  # R_sun
        M_star=1,  # M_sun
        time_span=1000,  # days
        period_min=0,
        period_max=999,
        oversampling_factor=3,
    )
    numpy.testing.assert_equal(len(periods), 4308558)
    print("passed")



from __future__ import division, print_function
import numpy
import batman
from foldedleastsquares import transitleastsquares


if __name__ == "__main__":
    print("Starting test: uncertainties...", end="")
    numpy.random.seed(seed=0)  # reproducibility

    # Create test data
    start = 48
    days = 365.25 * 3
    samples_per_day = 12  # 48
    samples = int(days * samples_per_day)  # 48
    t = numpy.linspace(start, start + days, samples)

    # Use batman to create transits
    ma = batman.TransitParams()
    ma.t0 = (
        start + 20
    )  # time of inferior conjunction; first transit is X days after start
    ma.per = 365.25  # orbital period
    ma.rp = 6371 / 696342  # 6371 planet radius (in units of stellar radii)
    ma.a = 217  # semi-major axis (in units of stellar radii)
    ma.inc = 90  # orbital inclination (in degrees)
    ma.ecc = 0  # eccentricity
    ma.w = 90  # longitude of periastron (in degrees)
    ma.u = [0.5]  # limb darkening coefficients
    ma.limb_dark = "linear"  # limb darkening model
    m = batman.TransitModel(ma, t)  # initializes model
    original_flux = m.light_curve(ma)  # calculates light curve

    # Create noise and merge with flux
    ppm = 5
    stdev = 10 ** -6 * ppm
    noise = numpy.random.normal(0, stdev, int(samples))
    y = original_flux + noise

    # Inject excess noise near the end of the time series.
    # When searching without uncertainties, the SDE is 4.827
    # When searching with uncertainties, the SDE is larger, 5.254. Test passed!
    noise = numpy.random.normal(0, 10 * stdev, 3149)
    y[10000:] = y[10000:] + noise

    dy = numpy.full(len(y), stdev)
    dy[10000:] = 10 * stdev

    model = transitleastsquares(t, y, dy)
    results = model.power(
        period_min=360,
        period_max=370,
        oversampling_factor=3,
        duration_grid_step=1.05,
        T0_fit_margin=0.2,
    )
    numpy.testing.assert_almost_equal(results.SDE, 5.292594615900944, decimal=5)
    print("passed")



from __future__ import division, print_function
import multiprocessing
import numpy
import sys
import warnings
from functools import partial
from tqdm import tqdm

# TLS parts
from .template_generator.tailed_transit_template_generator import TailedTransitTemplateGenerator
from .template_generator.default_transit_template_generator import DefaultTransitTemplateGenerator
from . import tls_constants
from .core import (
    search_period,
)
from .validate import validate_inputs, validate_args


class transitleastsquares(object):
    """Compute the transit least squares of limb-darkened transit models"""

    def __init__(self, t, y, dy=None):
        self.t, self.y, self.dy = validate_inputs(t, y, dy)
        default_transit_template_generator = DefaultTransitTemplateGenerator()
        self.transit_template_generators = {"default": default_transit_template_generator,
                               "grazing": default_transit_template_generator,
                               "box": default_transit_template_generator,
                               "tailed": TailedTransitTemplateGenerator()}

    def power(self, **kwargs):
        """Compute the periodogram for a set of user-defined parameters"""

        print(tls_constants.TLS_VERSION)
        self, kwargs = validate_args(self, kwargs)
        transit_template_generator = self.transit_template_generators[self.transit_template]
        periods = self.period_grid if self.period_grid is not None else transit_template_generator.period_grid(
            R_star=self.R_star,
            M_star=self.M_star,
            time_span=numpy.max(self.t) - numpy.min(self.t),
            period_min=self.period_min,
            period_max=self.period_max,
            oversampling_factor=self.oversampling_factor,
            n_transits_min=self.n_transits_min,
        )

        durations = transit_template_generator.duration_grid(
            periods, shortest=1 / len(self.t), log_step=self.duration_grid_step
        )

        maxwidth_in_samples = int(numpy.max(durations) * numpy.size(self.y))
        if maxwidth_in_samples % 2 != 0:
            maxwidth_in_samples = maxwidth_in_samples + 1
        lc_cache_overview, lc_arr = transit_template_generator.get_cache(
            period_grid=periods,
            durations=durations,
            maxwidth_in_samples=maxwidth_in_samples,
            per=self.per,
            rp=self.rp,
            a=self.a,
            inc=self.inc,
            ecc=self.ecc,
            w=self.w,
            u=self.u,
            limb_dark=self.limb_dark,
        )

        print(
            "Searching "
            + str(len(self.y))
            + " data points, "
            + str(len(periods))
            + " periods from "
            + str(round(min(periods), 3))
            + " to "
            + str(round(max(periods), 3))
            + " days"
        )

        # Python 2 multiprocessing with "partial" doesn't work
        # For now, only single-threading in Python 2 is supported
        if sys.version_info[0] < 3:
            self.use_threads = 1
            warnings.warn("This TLS version supports no multithreading on Python 2")

        if self.use_threads == multiprocessing.cpu_count():
            print("Using all " + str(self.use_threads) + " CPU threads")
        else:
            print(
                "Using "
                + str(self.use_threads)
                + " of "
                + str(multiprocessing.cpu_count())
                + " CPU threads"
            )

        if self.show_progress_bar:
            bar_format = "{desc}{percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} periods | {elapsed}<{remaining}"
            pbar = tqdm(total=numpy.size(periods), smoothing=0.3, bar_format=bar_format)

        if tls_constants.PERIODS_SEARCH_ORDER == "ascending":
            periods = reversed(periods)
        elif tls_constants.PERIODS_SEARCH_ORDER == "descending":
            pass  # it already is
        elif tls_constants.PERIODS_SEARCH_ORDER == "shuffled":
            periods = numpy.random.permutation(periods)
        else:
            raise ValueError("Unknown PERIODS_SEARCH_ORDER")

        # Result lists now (faster), convert to numpy array later
        test_statistic_periods = []
        test_statistic_residuals = []
        test_statistic_rows = []
        test_statistic_depths = []

        if self.use_threads > 1:  # Run multi-core search
            pool = multiprocessing.Pool(processes=self.use_threads)
            params = partial(
                search_period,
                transit_template_generator,
                t=self.t,
                y=self.y,
                dy=self.dy,
                transit_depth_min=self.transit_depth_min,
                R_star_min=self.R_star_min,
                R_star_max=self.R_star_max,
                M_star_min=self.M_star_min,
                M_star_max=self.M_star_max,
                lc_arr=lc_arr,
                lc_cache_overview=lc_cache_overview,
                T0_fit_margin=self.T0_fit_margin,
            )
            for data in pool.imap_unordered(params, periods):
                test_statistic_periods.append(data[0])
                test_statistic_residuals.append(data[1])
                test_statistic_rows.append(data[2])
                test_statistic_depths.append(data[3])
                if self.show_progress_bar:
                    pbar.update(1)
            pool.close()
        else:
            for period in periods:
                data = search_period(
                    transit_template_generator,
                    period=period,
                    t=self.t,
                    y=self.y,
                    dy=self.dy,
                    transit_depth_min=self.transit_depth_min,
                    R_star_min=self.R_star_min,
                    R_star_max=self.R_star_max,
                    M_star_min=self.M_star_min,
                    M_star_max=self.M_star_max,
                    lc_arr=lc_arr,
                    lc_cache_overview=lc_cache_overview,
                    T0_fit_margin=self.T0_fit_margin,
                )
                test_statistic_periods.append(data[0])
                test_statistic_residuals.append(data[1])
                test_statistic_rows.append(data[2])
                test_statistic_depths.append(data[3])
                if self.show_progress_bar:
                    pbar.update(1)

        if self.show_progress_bar:
            pbar.close()

        # imap_unordered delivers results in unsorted order ==> sort
        test_statistic_periods = numpy.array(test_statistic_periods)
        sort_index = numpy.argsort(test_statistic_periods)
        test_statistic_periods = test_statistic_periods[sort_index]
        test_statistic_residuals = numpy.array(test_statistic_residuals)[sort_index]
        test_statistic_rows = numpy.array(test_statistic_rows)[sort_index]
        test_statistic_depths = numpy.array(test_statistic_depths)[sort_index]

        idx_best = numpy.argmin(test_statistic_residuals)
        best_row = test_statistic_rows[idx_best]
        duration = lc_cache_overview["duration"][best_row]
        maxwidth_in_samples = int(numpy.max(durations) * numpy.size(self.t))

        if max(test_statistic_residuals) == min(test_statistic_residuals) and len(test_statistic_residuals) > 1:
            no_transits_were_fit = True
            warnings.warn('No transit were fit. Try smaller "transit_depth_min"')
        else:
            no_transits_were_fit = False

        # Power spectra variants
        chi2 = test_statistic_residuals
        degrees_of_freedom = 4
        chi2red = test_statistic_residuals / (len(self.t) - degrees_of_freedom)
        chi2_min = numpy.min(chi2)
        chi2red_min = numpy.min(chi2red)

        return transit_template_generator.calculate_results(no_transits_were_fit, chi2, chi2red, chi2_min,
                                                            chi2red_min, test_statistic_periods, test_statistic_depths,
                                                            self, lc_arr, best_row, periods,
                                                            durations, duration, maxwidth_in_samples, len(self.y))




from __future__ import division, print_function
import numpy
from foldedleastsquares import transitleastsquares

if __name__ == "__main__":

    t = y = numpy.linspace(0, 1, 2000)

    try:
        model = transitleastsquares(t, y)
        results = model.power(use_threads=0)
    except:
        print("Test passed: use_threads=0")
    try:
        model = transitleastsquares(t, y)
        results = model.power(use_threads="1")
    except:
        print("Test passed: use_threads='1'")



from __future__ import division, print_function
import numpy
from os import path
from . import tls_constants
from .helpers import running_median, transit_mask


def FAP(SDE):
    """Returns FAP (False Alarm Probability) for a given SDE"""
    data = numpy.genfromtxt(
        path.join(tls_constants.resources_dir, "fap.csv"),
        dtype="f8, f8",
        delimiter=",",
        names=["FAP", "SDE"],
    )
    return data["FAP"][numpy.argmax(data["SDE"] > SDE)]


def rp_rs_from_depth(depth, law, params):
    """Takes the maximum transit depth, limb-darkening law and parameters
    Returns R_P / R_S (ratio of planetary to stellar radius)
    Source: Heller 2019, https://arxiv.org/abs/1901.01730"""

    # Validations:
    # - LD law must exist
    # - All parameters must be floats or ints
    # - All parameters must be given in the correct quanitity for the law

    if len(params) == 1:
        params = float(params[0])

    if not isinstance(params, (float, int)) and not all(
        isinstance(x, (float, int)) for x in params
    ):
        raise ValueError("All limb-darkening parameters must be numbers")

    laws = "linear, quadratic, squareroot, logarithmic, nonlinear"
    if law not in laws:
        raise ValueError("Please provide a supported limb-darkening law:", laws)

    if law == "linear" and not isinstance(params, float):
        raise ValueError("Please provide exactly one parameter")

    if law in "quadratic, logarithmic, squareroot" and len(params) != 2:
        raise ValueError("Please provide exactly two limb-darkening parameters")

    if law == "nonlinear" and len(params) != 4:
        raise ValueError("Please provide exactly four limb-darkening parameters")

    # Actual calculations of the return value
    if law == "linear":
        return (depth * (1 - params / 3)) ** (1 / 2)

    if law == "quadratic":
        return (depth * (1 - params[0] / 3 - params[1] / 6)) ** (1 / 2)

    if law == "squareroot":
        return (depth * (1 - params[0] / 3 - params[1] / 5)) ** (1 / 2)

    if law == "logarithmic":
        return (depth * (1 + 2 * params[1] / 9 - params[0] / 3)) ** (1 / 2)

    if law == "nonlinear":
        return (
            depth
            * (1 - params[0] / 5 - params[1] / 3 - 3 * params[2] / 7 - params[3] / 2)
        ) ** (1 / 2)


def pink_noise(data, width):
    std = 0
    datapoints = len(data) - width + 1
    for i in range(datapoints):
        std += numpy.std(data[i : i + width]) / width ** 0.5
    return std / datapoints


def period_uncertainty(periods, power):
    # Determine estimate for uncertainty in period
    # Method: Full width at half maximum
    try:
        # Upper limit
        index_highest_power = numpy.argmax(power)
        idx = index_highest_power
        while True:
            idx += 1
            if power[idx] <= 0.5 * power[index_highest_power]:
                idx_upper = idx
                break
        # Lower limit
        idx = index_highest_power
        while True:
            idx -= 1
            if power[idx] <= 0.5 * power[index_highest_power]:
                idx_lower = idx
                break
        period_uncertainty = 0.5 * (periods[idx_upper] - periods[idx_lower])
    except:
        period_uncertainty = float("inf")
    return period_uncertainty


def spectra(chi2, oversampling_factor, chi2_baseline):
    SR = numpy.min(chi2) / chi2
    SDE_raw = (1 - numpy.mean(SR)) / numpy.std(SR)

    # Scale SDE_power from 0 to SDE_raw
    power_raw = SR - numpy.mean(SR)  # shift down to the mean being zero
    scale = SDE_raw / numpy.max(power_raw)  # scale factor to touch max=SDE_raw
    power_raw = power_raw * scale

    # Detrended SDE, named "power"
    kernel = int(oversampling_factor * tls_constants.SDE_MEDIAN_KERNEL_SIZE)
    if kernel % 2 == 0:
        kernel = kernel + 1
    if len(power_raw) > 2 * kernel:
        my_median = running_median(power_raw, kernel)
        power = power_raw - my_median
        # Re-normalize to range between median = 0 and peak = SDE
        # shift down to the mean being zero
        power = power - numpy.mean(power)
        power = power_neighbour_correction(power, kernel, chi2, chi2_baseline)
        SDE = numpy.max(power / numpy.std(power))
        # scale factor to touch max=SDE
        scale = SDE / numpy.max(power)
        power = power * scale
    else:
        power = power_raw
        SDE = SDE_raw

    return SR, power_raw, power, SDE_raw, SDE

def power_neighbour_correction(sde, kernel, chi2, chi2_baseline):
    """Returns scaled power spectra taking into account the neighbours used with the given kernel that are below the
    residuals baseline"""
    idx = numpy.arange(kernel) + numpy.arange(len(sde) - kernel + 1)[:, None]
    chi2_idx = chi2[idx]
    chi2_idx_gtbaseline = [len(numpy.argwhere(chi2_window < chi2_baseline).flatten()) for chi2_window in chi2_idx]
    # Append the first/last value at the beginning/end to match the length of
    # data and returned median
    first_values = chi2_idx_gtbaseline[0]
    last_values = chi2_idx_gtbaseline[-1]
    missing_values = len(sde) - len(chi2_idx_gtbaseline)
    values_front = int(missing_values * 0.5)
    values_end = missing_values - values_front
    chi2_idx_gtbaseline = numpy.append(numpy.full(values_front, first_values), chi2_idx_gtbaseline)
    chi2_idx_gtbaseline = numpy.append(chi2_idx_gtbaseline, numpy.full(values_end, last_values))
    sde = sde * chi2_idx_gtbaseline / kernel
    return sde


def model_lightcurve(transit_times, period, t, model_transit_single):
    """Creates the model light curve for the full unfolded dataset"""

    # Append one more transit after and before end of nominal time series
    # to fully cover beginning and end with out of transit calculations
    earlier_tt = transit_times[0] - period
    extended_transit_times = numpy.append(earlier_tt, transit_times)
    next_tt = transit_times[-1] + period
    extended_transit_times = numpy.append(extended_transit_times, next_tt)
    full_x_array = numpy.array([])
    full_y_array = numpy.array([])
    rounds = len(extended_transit_times)
    internal_samples = (
        int(len(t) / len(transit_times))
    ) * tls_constants.OVERSAMPLE_MODEL_LIGHT_CURVE

    # Append all periods
    for i in range(rounds):
        xmin = extended_transit_times[i] - period / 2
        xmax = extended_transit_times[i] + period / 2
        x_array = numpy.linspace(xmin, xmax, internal_samples)
        full_x_array = numpy.append(full_x_array, x_array)
        full_y_array = numpy.append(full_y_array, model_transit_single)

    if numpy.all(numpy.isnan(full_x_array)):
        return None, None
    else:  # Determine start and end of relevant time series, and crop it
        start_cadence = numpy.nanargmax(full_x_array > min(t))
        stop_cadence = numpy.nanargmax(full_x_array > max(t))
        full_x_array = full_x_array[start_cadence:stop_cadence]
        full_y_array = full_y_array[start_cadence:stop_cadence]
        model_lightcurve_model = full_y_array
        model_lightcurve_time = full_x_array
        return model_lightcurve_model, model_lightcurve_time


def all_transit_times(T0, t, period):
    """Return all mid-transit times within t"""

    if T0 < min(t):
        transit_times = [T0 + period]
    else:
        transit_times = [T0]
    previous_transit_time = transit_times[0]
    transit_number = 0
    while True:
        transit_number = transit_number + 1
        next_transit_time = previous_transit_time + period
        if next_transit_time < (numpy.min(t) + (numpy.max(t) - numpy.min(t))):
            transit_times.append(next_transit_time)
            previous_transit_time = next_transit_time
        else:
            break
    return transit_times


def calculate_transit_duration_in_days(t, period, transit_times, duration):
    """Return estimate for transit duration in days"""

    # Difference between (time series duration / period) and epochs
    transit_duration_in_days_raw = (
        duration * calculate_stretch(t, period, transit_times) * period
    )

    # Correct the duration for gaps in the data
    transit_duration_in_days = transit_duration_in_days_raw * calculate_fill_factor(t)

    return transit_duration_in_days


def calculate_stretch(t, period, transit_times):
    """Return difference between (time series duration / period) and epochs
        Example: 
        - Time series duration = 100 days
        - Period = 40 days
        - Epochs = 2 at t0s = [30, 70] days
        ==> stretch = (100 / 40) / 2 = 1.25"""

    duration_timeseries = (numpy.max(t) - numpy.min(t)) / period
    epochs = len(transit_times)
    stretch = duration_timeseries / epochs
    return stretch


def calculate_fill_factor(t):
    """Return the fraction of existing cadences, assuming constant cadences"""

    average_cadence = numpy.median(numpy.diff(t))
    span = max(t) - min(t)
    theoretical_cadences = span / average_cadence
    fill_factor = (len(t) - 1) / theoretical_cadences
    return fill_factor


def count_stats(t, y, transit_times, transit_duration_in_days):
    """Return:
    * in_transit_count:     Number of data points in transit (phase-folded)
    * after_transit_count:  Number of data points in a bin of transit duration, 
                            after transit (phase-folded)
    * before_transit_count: Number of data points in a bin of transit duration, 
                            before transit (phase-folded)
    """
    in_transit_count = 0
    after_transit_count = 0
    before_transit_count = 0

    for mid_transit in transit_times:
        T0 = (
            mid_transit - 1.5 * transit_duration_in_days
        )  # start of 1 transit dur before ingress
        T1 = mid_transit - 0.5 * transit_duration_in_days  # start of ingress
        T4 = mid_transit + 0.5 * transit_duration_in_days  # end of egress
        T5 = (
            mid_transit + 1.5 * transit_duration_in_days
        )  # end of egress + 1 transit dur

        if T0 > min(t) and T5 < max(t):  # inside time
            idx_intransit = numpy.where(numpy.logical_and(t > T1, t < T4))
            idx_before_transit = numpy.where(numpy.logical_and(t > T0, t < T1))
            idx_after_transit = numpy.where(numpy.logical_and(t > T4, t < T5))
            points_in_this_in_transit = len(y[idx_intransit])
            points_in_this_before_transit = len(y[idx_before_transit])
            points_in_this_after_transit = len(y[idx_after_transit])

            in_transit_count += points_in_this_in_transit
            before_transit_count += points_in_this_before_transit
            after_transit_count += points_in_this_after_transit

    return in_transit_count, after_transit_count, before_transit_count


def intransit_stats(t, y, transit_times, transit_duration_in_days):
    """Return all intransit odd and even flux points"""

    all_flux_intransit_odd = numpy.array([])
    all_flux_intransit_even = numpy.array([])
    all_flux_intransit = numpy.array([])
    all_idx_intransit = numpy.array([])
    per_transit_count = numpy.zeros([len(transit_times)])
    transit_depths = numpy.zeros([len(transit_times)])
    transit_depths_uncertainties = numpy.zeros([len(transit_times)])

    for i in range(len(transit_times)):

        depth_mean_odd = numpy.nan
        depth_mean_even = numpy.nan
        depth_mean_odd_std = numpy.nan
        depth_mean_even_std = numpy.nan

        mid_transit = transit_times[i]
        tmin = mid_transit - 0.5 * transit_duration_in_days
        tmax = mid_transit + 0.5 * transit_duration_in_days
        if numpy.isnan(tmin) or numpy.isnan(tmax):
            idx_intransit = []
            flux_intransit = []
            mean_flux = numpy.nan
        else:
            idx_intransit = numpy.where(numpy.logical_and(t > tmin, t < tmax))
            flux_intransit = y[idx_intransit]
            if len(y[idx_intransit]) > 0:
                mean_flux = numpy.mean(y[idx_intransit])
            else:
                mean_flux = numpy.nan
        intransit_points = numpy.size(y[idx_intransit])
        transit_depths[i] = mean_flux
        if len(y[idx_intransit] > 0):
            transit_depths_uncertainties[i] = numpy.std(y[idx_intransit]) / numpy.sqrt(
                intransit_points
            )
        else:
            transit_depths_uncertainties[i] = numpy.nan
        per_transit_count[i] = intransit_points

        # Check if transit odd/even to collect the flux for the mean calculations
        if i % 2 == 0:  # even
            all_flux_intransit_even = numpy.append(
                all_flux_intransit_even, flux_intransit
            )
        else:  # odd
            all_flux_intransit_odd = numpy.append(
                all_flux_intransit_odd, flux_intransit
            )
        if len(all_flux_intransit_odd) > 0:
            depth_mean_odd = numpy.mean(all_flux_intransit_odd)

            depth_mean_odd_std = numpy.std(all_flux_intransit_odd) / numpy.sum(
                len(all_flux_intransit_odd)
            ) ** (0.5)
        if len(all_flux_intransit_even) > 0:
            depth_mean_even = numpy.mean(all_flux_intransit_even)
            depth_mean_even_std = numpy.std(all_flux_intransit_even) / numpy.sum(
                len(all_flux_intransit_even)
            ) ** (0.5)

    return (
        depth_mean_odd,
        depth_mean_even,
        depth_mean_odd_std,
        depth_mean_even_std,
        all_flux_intransit_odd,
        all_flux_intransit_even,
        per_transit_count,
        transit_depths,
        transit_depths_uncertainties,
    )


def snr_stats(
    t,
    y,
    period,
    duration,
    T0,
    transit_times,
    transit_duration_in_days,
    per_transit_count,
    intransit=None
):
    """Return snr_per_transit and snr_pink_per_transit"""

    snr_per_transit = numpy.zeros([len(transit_times)])
    snr_pink_per_transit = numpy.zeros([len(transit_times)])
    if intransit is None:
        intransit = transit_mask(t, period, 2 * duration, T0)
    flux_ootr = y[~intransit]

    try:
        pinknoise = pink_noise(flux_ootr, int(numpy.mean(per_transit_count)))
    except:
        pinknoise = numpy.nan

    # Estimate SNR and pink SNR
    # Second run because now the out of transit points are known
    if len(flux_ootr) > 0:
        std = numpy.std(flux_ootr)
    else:
        std = numpy.nan
    for i in range(len(transit_times)):
        mid_transit = transit_times[i]
        tmin = mid_transit - 0.5 * transit_duration_in_days
        tmax = mid_transit + 0.5 * transit_duration_in_days
        if numpy.isnan(tmin) or numpy.isnan(tmax):
            idx_intransit = []
            mean_flux = numpy.nan
        else:
            idx_intransit = numpy.where(numpy.logical_and(t > tmin, t < tmax))
            if len(y[idx_intransit]) > 0:
                mean_flux = numpy.mean(y[idx_intransit])
            else:
                mean_flux = numpy.nan

        intransit_points = numpy.size(y[idx_intransit])
        try:
            snr_pink_per_transit[i] = (1 - mean_flux) / pinknoise if mean_flux < 1 else (mean_flux - 1) / pinknoise
            if intransit_points > 0 and not numpy.isnan(std):
                std_binned = std / intransit_points ** 0.5
                snr_per_transit[i] = (1 - mean_flux) / std_binned if mean_flux < 1 else (mean_flux - 1) / std_binned
            else:
                snr_per_transit[i] = 0
                snr_pink_per_transit[i] = 0
        except:
            snr_per_transit[i] = 0
            snr_pink_per_transit[i] = 0

    return snr_per_transit, snr_pink_per_transit



from __future__ import division, print_function
import numpy
import batman
from foldedleastsquares import transitleastsquares


# Test case for statistics when data contains gap during a transit

if __name__ == "__main__":
    print("Starting test: Gap in test statistic...", end="")
    numpy.random.seed(seed=0)  # reproducibility
    # Create test data
    start = 48
    days = 365.25 * 3
    samples_per_day = 12  # 48
    samples = int(days * samples_per_day)  # 48
    t = numpy.linspace(start, start + days, samples)

    # Use batman to create transits
    ma = batman.TransitParams()
    ma.t0 = start + 20
    ma.per = 365.25  # orbital period
    ma.rp = 6371 / 696342  # 6371 planet radius (in units of stellar radii)
    ma.a = 217  # semi-major axis (in units of stellar radii)
    ma.inc = 90  # orbital inclination (in degrees)
    ma.ecc = 0  # eccentricity
    ma.w = 90  # longitude of periastron (in degrees)
    ma.u = [0.5]  # limb darkening coefficients
    ma.limb_dark = "linear"  # limb darkening model
    m = batman.TransitModel(ma, t)  # initializes model
    original_flux = m.light_curve(ma)  # calculates light curve

    # Create noise and merge with flux
    ppm = 5
    stdev = 10 ** -6 * ppm
    noise = numpy.random.normal(0, stdev, int(samples))
    y = original_flux + noise
    y[1] = numpy.nan

    # print(len(y))
    y[200:500] = numpy.nan
    t[200:500] = numpy.nan
    # import matplotlib.pyplot as plt
    # plt.plot(t, y)
    # plt.show()

    model = transitleastsquares(t, y)
    results = model.power(
        period_min=360,
        period_max=370,
        transit_depth_min=10 * 10 ** -6,
        oversampling_factor=2,
        duration_grid_step=1.1,
        T0_fit_margin=1.2,
    )

    numpy.testing.assert_almost_equal(
        results.period_uncertainty, 0.3153203546531813, decimal=5
    )
    numpy.testing.assert_equal(results.per_transit_count, [0, 4, 4])
    numpy.testing.assert_equal(len(results.transit_times), 3)
    numpy.testing.assert_almost_equal(results.period, 365.22218620040417, decimal=5)
    numpy.testing.assert_almost_equal(
        results.transit_times,
        [68.08637, 433.30855, 798.53074],
        decimal=5,
    )

    numpy.testing.assert_almost_equal(results.depth, 0.9998972750356973, decimal=5)
    numpy.testing.assert_almost_equal(results.duration, 0.3695717615478269, decimal=5)
    numpy.testing.assert_almost_equal(results.SDE, 4.243572802600693, decimal=3)
    numpy.testing.assert_almost_equal(
        results.odd_even_mismatch, 0.4359025761888081, decimal=3
    )
    numpy.testing.assert_almost_equal(results.rp_rs, 0.009114758081257387, decimal=3)

    # Full light curve model
    numpy.testing.assert_almost_equal(
        numpy.sum(results.model_lightcurve_time), 38275494.19583159, decimal=3
    )
    numpy.testing.assert_almost_equal(
        numpy.sum(results.model_lightcurve_model), 64233.9941755991, decimal=3
    )

    transit_times_expected = [68.086, 433.309, 798.531]
    numpy.testing.assert_almost_equal(
        results.transit_times, transit_times_expected, decimal=3
    )
    numpy.testing.assert_almost_equal(results.duration, 0.3695717615478269, decimal=3)

    numpy.testing.assert_almost_equal(
        max(results.model_folded_phase), 1.0000380285975052, decimal=3
    )
    numpy.testing.assert_almost_equal(
        min(results.model_folded_phase), 3.8028597505324e-05, decimal=3
    )
    numpy.testing.assert_almost_equal(
        numpy.mean(results.model_folded_phase), 0.5000380285975052, decimal=3
    )

    numpy.testing.assert_almost_equal(
        results.depth_mean_even, (0.999915, 6.785539e-06), decimal=3
    )
    numpy.testing.assert_almost_equal(
        results.depth_mean_odd, (0.999920, 1.209993e-05), decimal=3
    )
    numpy.testing.assert_almost_equal(
        results.depth_mean, (0.999917, 6.086923e-06), decimal=3
    )

    numpy.testing.assert_almost_equal(
        results.transit_depths, [numpy.nan, 0.99991, 0.9999], decimal=3
    )

    numpy.testing.assert_almost_equal(
        results.transit_depths_uncertainties,
        [numpy.nan, 2.92371e-06, 4.48803e-06],
        decimal=3,
    )
    numpy.testing.assert_almost_equal(
        results.odd_even_mismatch, 0.4359025761888081, decimal=3
    )
    numpy.testing.assert_almost_equal(results.transit_count, 3, decimal=3)
    numpy.testing.assert_almost_equal(results.distinct_transit_count, 2, decimal=3)
    numpy.testing.assert_almost_equal(results.empty_transit_count, 1, decimal=3)
    numpy.testing.assert_almost_equal(
        results.snr_per_transit, [0., 32.666, 33.999], decimal=3
    )
    numpy.testing.assert_almost_equal(results.snr, 47.138942974480806, decimal=3)
    numpy.testing.assert_almost_equal(
        results.snr_pink_per_transit, [0., 47.049, 48.969], decimal=3
    )
    print("passed")



html_show_sourcelink = False
project = 'Transit Least Squares'
copyright = '2019 Michael Hippke, Rene Heller'
author = 'Michael Hippke, Rene Heller'
version = ''
release = ''
extensions = ['sphinx.ext.mathjax',]
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
language = None
exclude_patterns = []
pygments_style = None
html_theme = "sphinx_rtd_theme"
html_static_path = ['_static']
htmlhelp_basename = 'TLSdoc'
latex_elements = {
}
latex_documents = [
    (master_doc, 'TLS.tex', 'TLS Documentation',
     'Michael Hippke, Rene Heller', 'manual'),
]
man_pages = [
    (master_doc, 'tls', 'TLS Documentation',
     [author], 1)
]
texinfo_documents = [
    (master_doc, 'TLS', 'TLS Documentation',
     author, 'TLS', 'One line description of project.',
     'Miscellaneous'),
]
epub_title = project
epub_exclude_files = ['search.html']


import warnings
from abc import ABC, abstractmethod

import numpy
from numpy import pi, sqrt
from .. import tls_constants
from ..interpolation import interp1d
from tqdm import tqdm
from ..core import fold
from ..results import transitleastsquaresresults
from ..stats import count_stats, snr_stats, model_lightcurve, calculate_stretch, \
    calculate_fill_factor, calculate_transit_duration_in_days, all_transit_times, spectra, intransit_stats, FAP, \
    period_uncertainty, rp_rs_from_depth


class TransitTemplateGenerator(ABC):
    """
    Root class to be used to implement transit shapes and their behaviours.
    """
    def __init__(self):
        pass

    @abstractmethod
    def reference_transit(self, period_grid, duration_grid, samples, per, rp, a, inc, ecc, w, u, limb_dark):
        """
        Creates a reference transit with the desired shape
        :param period_grid: The grid of periods to be processed
        :param duration_grid: The grid of durations to be processed
        :param samples: The samples count
        :param per: The period for the template
        :param rp: The radius of the planet causing the transit. This only applies to some templates and is kept to
        respect the original TLS implementation.
        :param a: The semimajor axis of the planet causing the transit. This only applies to some templates and is kept to
        respect the original TLS implementation.
        :param inc: The inclination of the planet causing the transit. This only applies to some templates and is kept to
        respect the original TLS implementation.
        :param ecc: The eccentricity of the planet causing the transit. This only applies to some templates and is kept to
        respect the original TLS implementation.
        :param w:
        :param u:
        :param limb_dark: The limb darkening applied to the transit. This only applies to some templates and is kept to
        respect the original TLS implementation.
        """
        pass

    @abstractmethod
    def duration_grid(self, periods, shortest, log_step=tls_constants.DURATION_GRID_STEP):
        """
        Generates a grid of durations.
        :param periods: The grid of periods to be used.
        :param shortest: The shortest duration.
        :param log_step: The logarithmic step increment factor.
        """
        pass

    def period_grid(self, R_star, M_star, time_span, period_min=0, period_max=float("inf"),
            oversampling_factor=tls_constants.OVERSAMPLING_FACTOR, n_transits_min=tls_constants.N_TRANSITS_MIN,
                    effective_time_span=None):
        """
        Generates a grid of optimal sampling periods for transit search in light curves.
        Following Ofir (2014, A&A, 561, A138)
        :param R_star: The star radius
        :param M_star: The star mass
        :param time_span: The light curve timespam
        :param period_min: The minimum period of the grid
        :param period_max: The maximum period of the grid
        :param oversampling_factor: The grid density free parameter
        :param n_transits_min: The minimum number of transits to be matched
        """
        if R_star < 0.1:
            text = (
                    "Warning: R_star was set to 0.1 for period_grid (was unphysical: "
                    + str(R_star)
                    + ")"
            )
            warnings.warn(text)
            R_star = 0.1

        if R_star > 10000:
            text = (
                    "Warning: R_star was set to 10000 for period_grid (was unphysical: "
                    + str(R_star)
                    + ")"
            )
            warnings.warn(text)
            R_star = 10000

        if M_star < 0.01:
            text = (
                    "Warning: M_star was set to 0.01 for period_grid (was unphysical: "
                    + str(M_star)
                    + ")"
            )
            warnings.warn(text)
            M_star = 0.01

        if M_star > 1000:
            text = (
                    "Warning: M_star was set to 1000 for period_grid (was unphysical: "
                    + str(M_star)
                    + ")"
            )
            warnings.warn(text)
            M_star = 1000
        if effective_time_span is None:
            effective_time_span = time_span
        R_star = R_star * tls_constants.R_sun
        M_star = M_star * tls_constants.M_sun
        time_span = time_span * tls_constants.SECONDS_PER_DAY  # seconds
        effective_time_span = effective_time_span * tls_constants.SECONDS_PER_DAY
        # boundary conditions
        f_min = n_transits_min / time_span
        f_max = 1.0 / (2 * pi) * sqrt(tls_constants.G * M_star / (3 * R_star) ** 3)
        P_min = 1 / f_max / tls_constants.SECONDS_PER_DAY
        if 0 < period_min < P_min:
            warnings.warn("You selected a `period_min` shorter than the Roche Limit period of " + str(numpy.round(P_min, 2)))
        # optimal frequency sampling, Equations (5), (6), (7)
        A = (
                (2 * pi) ** (2.0 / 3)
                / pi
                * R_star
                / (tls_constants.G * M_star) ** (1.0 / 3)
                / (effective_time_span * oversampling_factor)
        )
        C = f_min ** (1.0 / 3) - A / 3.0
        N_opt = (f_max ** (1.0 / 3) - f_min ** (1.0 / 3) + A / 3) * 3 / A

        X = numpy.arange(N_opt) + 1
        f_x = (A / 3 * X + C) ** 3
        P_x = 1 / f_x

        # Cut to given (optional) selection of periods
        periods = P_x / tls_constants.SECONDS_PER_DAY
        selected_index = numpy.where(
            numpy.logical_and(periods > period_min, periods <= period_max)
        )

        number_of_periods = numpy.size(periods[selected_index])

        if number_of_periods > 10 ** 6:
            text = (
                    "period_grid generates a very large grid ("
                    + str(number_of_periods)
                    + "). Recommend to check physical plausibility for stellar mass, radius, and time series duration."
            )
            warnings.warn(text)

        if number_of_periods < tls_constants.MINIMUM_PERIOD_GRID_SIZE:
            if time_span < 5 * tls_constants.SECONDS_PER_DAY:
                time_span = 5 * tls_constants.SECONDS_PER_DAY
            warnings.warn(
                "period_grid defaults to R_star=1 and M_star=1 as given density yielded grid with too few values"
            )
            return self.period_grid(
                R_star=1, M_star=1, time_span=time_span / tls_constants.SECONDS_PER_DAY
            )
        else:
            return periods[selected_index]  # periods in [days]

    @abstractmethod
    def min_duration(self, period, R_star, M_star, periods=None):
        """
        Calculates the minimum duration for this template.
        :param period: The period for which the duration needs to be calculated.
        :param R_star: The radius of the host star.
        :param M_star: The mass of the host star
        :param periods: The period grid.
        """
        pass

    @abstractmethod
    def max_duration(self, period, R_star, M_star, periods=None):
        """
        Calculates the maximum duration for this template.
        :param period: The period for which the duration needs to be calculated.
        :param R_star: The radius of the host star.
        :param M_star: The mass of the host star
        :param periods: The period grid.
        """
        pass

    @abstractmethod
    def final_T0_fit(self, signal, depth, t, y, dy, period, T0_fit_margin, show_progress_bar):
        """ After the search, we know the best period, width and duration.
            But T0 was not preserved due to speed optimizations.
            Thus, iterate over T0s using the given parameters
            Fold to all T0s so that the transit is expected at phase = 0"""
        pass

    @abstractmethod
    def transit_mask(self, t, period, duration, T0):
        """
        Computes a boolean mask for the entire time set given by 't' and the transit parameters 'duration', 'T0' and
        'period'.
        :param t: the time set
        :param period: the transit period
        :param duration: the transit duration
        :param T0: the transit T0
        :return: the boolean transit mask
        """
        # Works with numba, but is not faster
        mask = numpy.abs((t - T0 + 0.5 * period) % period - 0.5 * period) < 0.5 * duration
        return mask

    @abstractmethod
    def calculate_results(self, no_transits_were_fit, chi2, chi2red, chi2_min, chi2red_min, test_statistic_periods,
                          test_statistic_depths, transitleastsquares, lc_arr, best_row, period_grid, durations,
                          duration, maxwidth_in_samples, chi2_baseline):
        """
        Returns a transitleastsquaresresult for the given computed statistics.
        """
        pass

    def fractional_transit(self,
            period_grid,
            duration_grid,
            duration,
            maxwidth,
            depth,
            samples,
            per,
            rp,
            a,
            inc,
            ecc,
            w,
            u,
            limb_dark,
            cached_reference_transit=None,
    ):
        """Returns a scaled reference transit with fractional width and depth"""

        if cached_reference_transit is None:
            reference_flux = self.reference_transit(
                period_grid=period_grid,
                duration_grid=duration_grid,
                samples=samples,
                per=per,
                rp=rp,
                a=a,
                inc=inc,
                ecc=ecc,
                w=w,
                u=u,
                limb_dark=limb_dark,
            )
        else:
            reference_flux = cached_reference_transit

        # Interpolate to shorter interval - new method without scipy
        reference_time = numpy.linspace(-0.5, 0.5, samples)
        occupied_samples = int((duration / maxwidth) * samples)
        x_new = numpy.linspace(-0.5, 0.5, occupied_samples)
        f = interp1d(x_new, reference_time)
        y_new = f(reference_flux)

        # Patch ends with ones ("1")
        missing_samples = samples - occupied_samples
        emtpy_segment = numpy.ones(int(missing_samples * 0.5))
        result = numpy.append(emtpy_segment, y_new)
        result = numpy.append(result, emtpy_segment)
        if numpy.size(result) < samples:  # If odd number of samples
            result = numpy.append(result, numpy.ones(1))

        # Depth rescaling
        result = 1 - ((1 - result) * depth)

        return result

    def get_cache(self, period_grid, durations, maxwidth_in_samples, per, rp, a, inc, ecc, w, u, limb_dark):
        """Fetches (size(durations)*size(depths)) light curves of length
            maxwidth_in_samples and returns these LCs in a 2D array, together with
            their metadata in a separate array."""

        print("Creating model cache for", str(len(durations)), "durations")
        lc_arr = []
        rows = numpy.size(durations)
        lc_cache_overview = numpy.zeros(
            rows,
            dtype=[("duration", "f8"), ("width_in_samples", "i8"), ("overshoot", "f8")],
        )
        cached_reference_transit = self.reference_transit(
            period_grid=period_grid,
            duration_grid=durations,
            samples=maxwidth_in_samples,
            per=per,
            rp=rp,
            a=a,
            inc=inc,
            ecc=ecc,
            w=w,
            u=u,
            limb_dark=limb_dark,
        )

        row = 0
        for duration in durations:
            scaled_transit = self.fractional_transit(
                period_grid=period_grid,
                duration_grid=durations,
                duration=duration,
                maxwidth=numpy.max(durations),
                depth=tls_constants.SIGNAL_DEPTH,
                samples=maxwidth_in_samples,
                per=per,
                rp=rp,
                a=a,
                inc=inc,
                ecc=ecc,
                w=w,
                u=u,
                limb_dark=limb_dark,
                cached_reference_transit=cached_reference_transit,
            )
            lc_cache_overview["duration"][row] = duration
            used_samples = int((duration / numpy.max(durations)) * maxwidth_in_samples)
            lc_cache_overview["width_in_samples"][row] = used_samples
            full_values = numpy.where(
                scaled_transit < (1 - tls_constants.NUMERICAL_STABILITY_CUTOFF)
            )
            first_sample = numpy.min(full_values)
            last_sample = numpy.max(full_values) + 1
            signal = scaled_transit[first_sample:last_sample]
            lc_arr.append(signal)

            # Fraction of transit bottom and mean flux
            overshoot = numpy.mean(signal) / numpy.min(signal)

            # Later, we multiply the inverse fraction ==> convert to inverse percentage
            lc_cache_overview["overshoot"][row] = 1 / (2 - overshoot)
            row += +1

        lc_arr = numpy.array(lc_arr, dtype=object)
        return lc_cache_overview, lc_arr


from __future__ import division, print_function
import numpy
from foldedleastsquares import FAP

if __name__ == "__main__":
    print("Starting test FAP...", end="")
    numpy.testing.assert_equal(FAP(SDE=2), numpy.nan)
    numpy.testing.assert_equal(FAP(SDE=7), 0.009443778)
    numpy.testing.assert_equal(FAP(SDE=99), 8.0032e-05)
    print("passed")



from __future__ import division, print_function
import numpy
import scipy
import scipy.signal
from foldedleastsquares import transitleastsquares, transit_mask, cleaned_array


def loadfile(filename):
    data = numpy.genfromtxt(filename, delimiter=",", dtype="f8, f8", names=["t", "y"])
    return data["t"], data["y"]


if __name__ == "__main__":
    print("Starting test: Multi-planet...", end="")
    t, y = loadfile("EPIC201367065.csv")
    trend = scipy.signal.medfilt(y, 25)
    y_filt = y / trend

    model = transitleastsquares(t, y_filt)
    results = model.power()

    numpy.testing.assert_almost_equal(max(results.power), 47.585106066175584, decimal=3)
    numpy.testing.assert_almost_equal(
        max(results.power_raw), 42.93056655774114, decimal=3
    )
    numpy.testing.assert_almost_equal(min(results.power), -0.6531152137309356, decimal=3)
    numpy.testing.assert_almost_equal(
        min(results.power_raw), -0.3043720539933344, decimal=3
    )
    print("Detrending of power spectrum from power_raw passed")

    # Mask of the first planet
    intransit = transit_mask(t, results.period, 2 * results.duration, results.T0)
    y_second_run = y_filt[~intransit]
    t_second_run = t[~intransit]
    t_second_run, y_second_run = cleaned_array(t_second_run, y_second_run)

    # Search for second planet
    model_second_run = transitleastsquares(t_second_run, y_second_run)
    results_second_run = model_second_run.power()
    numpy.testing.assert_almost_equal(
        results_second_run.duration, 0.14596011421893682, decimal=3
    )
    numpy.testing.assert_almost_equal(
        results_second_run.SDE, 35.031824450265, decimal=3
    )
    numpy.testing.assert_almost_equal(
        results_second_run.rp_rs, 0.025852178872027086, decimal=3
    )

    print("Passed")



from __future__ import division, print_function
import sys
from os import path
from . import version as tls_version

"""Magic constants"""
resources_dir = path.join(path.dirname(__file__))
TLS_VERSION = (
    "Transit Least Squares TLS "
    + tls_version.TLS_VERSIONING
    + " ("
    + tls_version.TLS_DATE
    + ")"
)

# astrophysical constants
G = 6.673e-11  # gravitational constant [m^3 / kg / s^2]
R_sun = 695508000  # radius of the Sun [m]
R_earth = 6371000  # radius of the Earth [m]
R_jup = 69911000  # radius of Jupiter [m]
M_sun = 1.989 * 10 ** 30  # mass of the Sun [kg]
SECONDS_PER_DAY = 86400

# Default values as described in the paper
TRANSIT_DEPTH_MIN = 10 * 10 ** -6  # 10 ppm
NUMERICAL_STABILITY_CUTOFF = 0.01 * 10 ** -6  # to identify nominal flux in model

# For the period grid
R_STAR = 1.0
M_STAR = 1.0
OVERSAMPLING_FACTOR = 3
N_TRANSITS_MIN = 2

# For the duration grid
M_STAR_MIN = 0.1
M_STAR_MAX = 1.0
R_STAR_MIN = 0.13
R_STAR_MAX = 3.5
DURATION_GRID_STEP = 1.1

# For the transit template
# quadratic limb darkening for a G2V star in the Kepler bandpass
# http://vizier.u-strasbg.fr/viz-bin/VizieR?-source=J/A%2BA/552/A16
DEFAULT_U = [0.4804, 0.1867]
DEFAULT_LIMB_DARK = "quadratic"
DEFAULT_ECC = 0  # eccentricity
DEFAULT_W = 90  # longitude of periastron (in degrees)
DEFAULT_PERIOD = 12.9
DEFAULT_RP = 0.03
DEFAULT_A = 23.1
DEFAULT_INC = 89.21

# Grazing template
GRAZING_B = 0.99

# Box-shaped (steep trapezoid) template
BOX_PERIOD = 29
BOX_RP = 0.1
BOX_A = 26.9
BOX_B = 0
BOX_INC = 90
BOX_U = [0]
BOX_LIMB_DARK = "linear"

# Unique depth of trial signals (at various durations). These are rescaled in
# depth so that their integral matches the mean flux in the window in question.
# In principle, "signal_depth" is an arbitrary value >0 and <1
SIGNAL_DEPTH = 0.5

# Maximum fractional transit duration ever observed is 0.117
# for Kepler-1368 b (as of Oct 2018), so we set upper_limit=0.15
# Long fractional transit durations are computationally expensive
# following a quadratic relation. If applicable, use a different value.
# Longer transits can still be found, but at decreasing sensitivity
FRACTIONAL_TRANSIT_DURATION_MAX = 0.12

# Oversampling ==> Downsampling of the reference transit:
# "Do not fit an unbinned model to binned data."
# Reference: Kipping, D., "Binning is sinning: morphological light-curve
#            distortions due to finite integration time"
#            MNRAS, Volume 408, Issue 3, pp. 1758-1769
#            http://adsabs.harvard.edu/cgi-bin/bib_query?arXiv:1004.3741
# This is not time-critical as it has to be done only once
# To work for all periods correctly, it has to be re-done at each period
# This feature is currently not implemented
SUPERSAMPLE_SIZE = 10000
OVERSAMPLE_MODEL_LIGHT_CURVE = 5

# Order in which the periods are searched: "shuffled"", "descending", "ascending"
# Shuffled has the advantage of the best estimate for the remaining time
PERIODS_SEARCH_ORDER = "shuffled"

# When converting power_raw to power, a median of a certain window size is subtracted.
# For periodograms of smaller width, no smoothing is applied. The kernel size is
# calculated as kernel = oversampling_factor * SDE_MEDIAN_KERNEL_SIZE
# This value has proven to yield numerically stable results.
SDE_MEDIAN_KERNEL_SIZE = 30

# AFFECTS ONLY THE FINAL T0 FIT, NOT THE SDE
# We can give user the option to not scan the phase space for T0 at every cadence
# For speed reasons, it may be acceptable to approximate T0 to within X %
# Useful in large datasets. 100k points: Extra runtime of order 2 minutes
# While individual transits often are only a few cadences long, in the stacked
# phase space it is (N transits * transit duration) [cadences] long
T0_FIT_MARGIN = 0.01  # of transit duration e.g., 0.01 (=1%)

# The secondary fit for T0 can take negligible or significant time, depending on the
# number of data points and on T0_FIT_MARGIN. Set an empirical threshold to avoid
# displaying a progress bar when the estimated runtime is low (<~1 sec)
# To display the progress bar in more cases, use a lower number
PROGRESSBAR_THRESHOLD = 5000

# Some cases yield empty period grids. Example: R_star=5, M_star=1
# Then: Warn and return the default grid
MINIMUM_PERIOD_GRID_SIZE = 100

# Warn the user if unknown **kwargs are given as parameters
VALID_PARAMETERS = [
    "R_star",
    "R_star_min",
    "R_star_max",
    "M_star",
    "M_star_min",
    "M_star_max",
    "period_min",
    "period_max",
    "n_transits_min",
    "per",
    "rp",
    "a",
    "inc",
    "b",
    "ecc",
    "w",
    "u",
    "limb_dark",
    "duration_grid_step",
    "transit_depth_min",
    "oversampling_factor",
    "T0_fit_margin",
    "use_threads",
    "show_progress_bar",
    "transit_template",
    "transit_template_generator",
    "period_grid"
]



import batman
import numpy
import warnings
from .. import tls_constants
from ..grid import T14
from ..interpolation import interp1d
from .transit_template_generator import TransitTemplateGenerator
from tqdm import tqdm
from ..core import fold
from ..results import transitleastsquaresresults
from ..stats import count_stats, snr_stats, model_lightcurve, calculate_stretch, \
    calculate_fill_factor, calculate_transit_duration_in_days, all_transit_times, spectra, intransit_stats, FAP, \
    period_uncertainty, rp_rs_from_depth


class DefaultTransitTemplateGenerator(TransitTemplateGenerator):
    """
    Default implementation used by TLS.
    """
    def __init__(self):
        super().__init__()

    def reference_transit(self, period_grid, duration_grid, samples, per, rp, a, inc, ecc, w, u, limb_dark):
        f = numpy.ones(tls_constants.SUPERSAMPLE_SIZE)
        duration = 1  # transit duration in days. Increase for exotic cases
        t = numpy.linspace(-duration * 0.5, duration * 0.5, tls_constants.SUPERSAMPLE_SIZE)
        ma = batman.TransitParams()
        ma.t0 = 0  # time of inferior conjunction
        ma.per = per  # orbital period, use Earth as a reference
        ma.rp = rp  # planet radius (in units of stellar radii)
        ma.a = a  # semi-major axis (in units of stellar radii)
        ma.inc = inc  # orbital inclination (in degrees)
        ma.ecc = ecc  # eccentricity
        ma.w = w  # longitude of periastron (in degrees)
        ma.u = u  # limb darkening coefficients
        ma.limb_dark = limb_dark  # limb darkening model
        m = batman.TransitModel(ma, t)  # initializes model
        flux = m.light_curve(ma)  # calculates light curve
        # Determine start of transit (first value < 1)
        idx_first = numpy.argmax(flux < 1)
        intransit_time = t[idx_first: -idx_first + 1]
        intransit_flux = flux[idx_first: -idx_first + 1]
        # Downsample (bin) to target sample size
        x_new = numpy.linspace(t[idx_first], t[-idx_first - 1], samples, per)
        f = interp1d(x_new, intransit_time)
        downsampled_intransit_flux = f(intransit_flux)
        # Rescale to height [0..1]
        rescaled = (numpy.min(downsampled_intransit_flux) - downsampled_intransit_flux) / (
                numpy.min(downsampled_intransit_flux) - 1
        )
        return rescaled

    def duration_grid(self, periods, shortest, log_step=tls_constants.DURATION_GRID_STEP):
        duration_max = self.max_duration(min(periods), tls_constants.R_STAR_MAX, tls_constants.M_STAR_MAX)
        duration_min = self.min_duration(max(periods), tls_constants.R_STAR_MIN, tls_constants.M_STAR_MIN)
        durations = [duration_min]
        current_depth = duration_min
        while current_depth * log_step < duration_max:
            current_depth = current_depth * log_step
            durations.append(current_depth)
        durations.append(duration_max)  # Append endpoint. Not perfectly spaced.
        return durations

    def min_duration(self, period, R_star, M_star, periods=None):
        return T14(R_s=R_star, M_s=M_star, P=period, small=True)

    def max_duration(self, period, R_star, M_star, periods=None):
        return T14(R_s=R_star, M_s=M_star, P=period, small=False)

    def final_T0_fit(self, signal, depth, t, y, dy, period, T0_fit_margin, show_progress_bar):
        dur = len(signal)
        scale = tls_constants.SIGNAL_DEPTH / (1 - depth) if depth >= 0 else tls_constants.SIGNAL_DEPTH / (1 + depth)
        signal = [1 - ((1 - value) / scale) if value <= 1 else 1 + ((value - 1) / scale) for value in signal]
        samples_per_period = numpy.size(y)

        if T0_fit_margin == 0:
            points = samples_per_period
        else:
            step_factor = T0_fit_margin * dur
            points = int(samples_per_period / step_factor)
        if points > samples_per_period:
            points = samples_per_period

        # Create all possible T0s from the start of [t] to [t+period] in [samples] steps
        T0_array = numpy.linspace(
            start=numpy.min(t), stop=numpy.min(t) + period, num=points
        )

        # Avoid showing progress bar when expected runtime is short
        if points > tls_constants.PROGRESSBAR_THRESHOLD and show_progress_bar:
            show_progress_info = True
        else:
            show_progress_info = False

        residuals_lowest = float("inf")
        T0 = 0

        if show_progress_info:
            print("Searching for best T0 for period", format(period, ".5f"), "days")
            pbar2 = tqdm(total=numpy.size(T0_array))
        signal_ootr = numpy.ones(len(y[dur:]))

        # Future speed improvement possible: Add multiprocessing. Will be slower for
        # short data and T0_FIT_MARGIN > 0.01, but faster for large data with dense
        # sampling (T0_FIT_MARGIN=0)
        for Tx in T0_array:
            phases = fold(time=t, period=period, T0=Tx)
            sort_index = numpy.argsort(phases, kind="mergesort")  # 75% of CPU time
            phases = phases[sort_index]
            flux = y[sort_index]
            dy = dy[sort_index]

            # Roll so that the signal starts at index 0
            # Numpy roll is slow, so we replace it with less elegant concatenate
            # flux = numpy.roll(flux, roll_cadences)
            # dy = numpy.roll(dy, roll_cadences)
            roll_cadences = int(dur / 2) + 1
            flux = numpy.concatenate([flux[-roll_cadences:], flux[:-roll_cadences]])
            dy = numpy.concatenate([flux[-roll_cadences:], flux[:-roll_cadences]])

            residuals_intransit = numpy.sum((flux[:dur] - signal) ** 2 / dy[:dur] ** 2)
            residuals_ootr = numpy.sum((flux[dur:] - signal_ootr) ** 2 / dy[dur:] ** 2)
            residuals_total = residuals_intransit + residuals_ootr

            if show_progress_info:
                pbar2.update(1)
            if residuals_total < residuals_lowest:
                residuals_lowest = residuals_total
                T0 = Tx
        if show_progress_info:
            pbar2.close()
        return T0

    def transit_mask(self, t, period, duration, T0):
        # Works with numba, but is not faster
        mask = numpy.abs((t - T0 + 0.5 * period) % period - 0.5 * period) < 0.5 * duration
        return mask

    def calculate_results(self, no_transits_were_fit, chi2, chi2red, chi2_min, chi2red_min, test_statistic_periods,
                          test_statistic_depths, transitleastsquares, lc_arr, best_row, period_grid, durations,
                          duration, maxwidth_in_samples, chi2_baseline):
        """
        Returns a transitleastsquaresresult for the given template
        """
        if no_transits_were_fit:
            power_raw = numpy.zeros(len(chi2))
            power = numpy.zeros(len(chi2))
            period = numpy.nan
            depth = 1
            SR = 0
            SDE = 0
            SDE_raw = 0
            T0 = 0
            transit_times = numpy.nan
            transit_duration_in_days = numpy.nan
            internal_samples = (
                int(len(transitleastsquares.y)) * tls_constants.OVERSAMPLE_MODEL_LIGHT_CURVE
            )
            folded_phase = numpy.nan
            folded_y = numpy.nan
            folded_dy = numpy.nan
            model_folded_phase = numpy.nan
            model_folded_model = numpy.nan
            model_transit_single = numpy.nan
            model_lightcurve_model = numpy.nan
            model_lightcurve_time = numpy.nan
            depth_mean_odd = numpy.nan
            depth_mean_even = numpy.nan
            depth_mean_odd_std = numpy.nan
            depth_mean_even_std = numpy.nan
            all_flux_intransit_odd = numpy.nan
            all_flux_intransit_even = numpy.nan
            per_transit_count = numpy.nan
            transit_depths = numpy.nan
            transit_depths_uncertainties = numpy.nan
            all_flux_intransit = numpy.nan
            snr_per_transit = numpy.nan
            snr_pink_per_transit = numpy.nan
            depth_mean = numpy.nan
            depth_mean_std = numpy.nan
            snr = numpy.nan
            rp_rs = numpy.nan
            depth_mean_odd = numpy.nan
            depth_mean_even = numpy.nan
            depth_mean_odd_std = numpy.nan
            depth_mean_even_std = numpy.nan
            odd_even_difference = numpy.nan
            odd_even_std_sum = numpy.nan
            odd_even_mismatch = numpy.nan
            transit_count = numpy.nan
            empty_transit_count = numpy.nan
            distinct_transit_count = numpy.nan
            duration = numpy.nan
            in_transit_count = numpy.nan
            after_transit_count = numpy.nan
            before_transit_count = numpy.nan
        else:
            SR, power_raw, power, SDE_raw, SDE = spectra(chi2, transitleastsquares.oversampling_factor, chi2_baseline)
            index_highest_power = numpy.argmax(power)
            period = test_statistic_periods[index_highest_power]
            depth = test_statistic_depths[index_highest_power]
            T0 = self.final_T0_fit(
                signal=lc_arr[best_row],
                depth=depth,
                t=transitleastsquares.t,
                y=transitleastsquares.y,
                dy=transitleastsquares.dy,
                period=period,
                T0_fit_margin=transitleastsquares.T0_fit_margin,
                show_progress_bar=transitleastsquares.show_progress_bar,
            )
            transit_times = all_transit_times(T0, transitleastsquares.t, period)

            transit_duration_in_days = calculate_transit_duration_in_days(
                transitleastsquares.t, period, transit_times, duration
            )
            phases = fold(transitleastsquares.t, period, T0=T0 + period / 2)
            sort_index = numpy.argsort(phases)
            folded_phase = phases[sort_index]
            folded_y = transitleastsquares.y[sort_index]
            folded_dy = transitleastsquares.dy[sort_index]
            # Model phase, shifted by half a cadence so that mid-transit is at phase=0.5
            model_folded_phase = numpy.linspace(
                0 + 1 / numpy.size(transitleastsquares.t) / 2,
                1 + 1 / numpy.size(transitleastsquares.t) / 2,
                numpy.size(transitleastsquares.t),
            )
            # Folded model / model curve
            # Data phase 0.5 is not always at the midpoint (not at cadence: len(y)/2),
            # so we need to roll the model to match the model so that its mid-transit
            # is at phase=0.5
            fill_factor = calculate_fill_factor(transitleastsquares.t)
            fill_half = 1 - ((1 - fill_factor) * 0.5)
            stretch = calculate_stretch(transitleastsquares.t, period, transit_times)
            internal_samples = (
                int(len(transitleastsquares.y) / len(transit_times))
            ) * tls_constants.OVERSAMPLE_MODEL_LIGHT_CURVE

            # Folded model flux
            model_folded_model = self.fractional_transit(
                period_grid=period_grid,
                duration_grid=durations,
                duration=duration * maxwidth_in_samples * fill_half,
                maxwidth=maxwidth_in_samples / stretch,
                depth=1 - depth,
                samples=int(len(transitleastsquares.t / len(transit_times))),
                per=transitleastsquares.per,
                rp=transitleastsquares.rp,
                a=transitleastsquares.a,
                inc=transitleastsquares.inc,
                ecc=transitleastsquares.ecc,
                w=transitleastsquares.w,
                u=transitleastsquares.u,
                limb_dark=transitleastsquares.limb_dark,
            )
            intransit_folded_model = numpy.where(model_folded_model < 1.)[0]
            if len(intransit_folded_model) > 0:
                transit_duration_in_days = period * (model_folded_phase[intransit_folded_model[-1]] -
                                                     model_folded_phase[intransit_folded_model[0]])
            # Full unfolded light curve model
            model_transit_single = self.fractional_transit(
                period_grid=period_grid,
                duration_grid=durations,
                duration=(duration * maxwidth_in_samples),
                maxwidth=maxwidth_in_samples / stretch,
                depth=1 - depth,
                samples=internal_samples,
                per=transitleastsquares.per,
                rp=transitleastsquares.rp,
                a=transitleastsquares.a,
                inc=transitleastsquares.inc,
                ecc=transitleastsquares.ecc,
                w=transitleastsquares.w,
                u=transitleastsquares.u,
                limb_dark=transitleastsquares.limb_dark,
            )
            model_lightcurve_model, model_lightcurve_time = model_lightcurve(
                transit_times, period, transitleastsquares.t, model_transit_single
            )
            depth_mean_odd, depth_mean_even, depth_mean_odd_std, depth_mean_even_std, all_flux_intransit_odd, \
            all_flux_intransit_even, per_transit_count, transit_depths, transit_depths_uncertainties = intransit_stats(
                transitleastsquares.t, transitleastsquares.y, transit_times, transit_duration_in_days
            )
            all_flux_intransit = numpy.concatenate(
                [all_flux_intransit_odd, all_flux_intransit_even]
            )
            snr_per_transit, snr_pink_per_transit = snr_stats(
                t=transitleastsquares.t,
                y=transitleastsquares.y,
                period=period,
                duration=duration,
                T0=T0,
                transit_times=transit_times,
                transit_duration_in_days=transit_duration_in_days,
                per_transit_count=per_transit_count,
                intransit=self.transit_mask(transitleastsquares.t, period, 2 * duration, T0)
            )
            intransit = self.transit_mask(transitleastsquares.t, period, 2 * duration, T0)
            flux_ootr = transitleastsquares.y[~intransit]
            depth_mean = numpy.mean(all_flux_intransit)
            depth_mean_std = numpy.std(all_flux_intransit) / numpy.sum(
                per_transit_count
            ) ** (0.5)
            flux_depth_mean_diff = 1 - depth_mean if depth_mean < 1 else depth_mean - 1
            snr = (flux_depth_mean_diff / numpy.std(flux_ootr)) * len(
                all_flux_intransit
            ) ** (0.5)

            if len(all_flux_intransit_odd) > 0:
                depth_mean_odd = numpy.mean(all_flux_intransit_odd)
                depth_mean_odd_std = numpy.std(all_flux_intransit_odd) / numpy.sum(
                    len(all_flux_intransit_odd)
                ) ** (0.5)
            else:
                depth_mean_odd = numpy.nan
                depth_mean_odd_std = numpy.nan

            if len(all_flux_intransit_even) > 0:
                depth_mean_even = numpy.mean(all_flux_intransit_even)
                depth_mean_even_std = numpy.std(all_flux_intransit_even) / numpy.sum(
                    len(all_flux_intransit_even)
                ) ** (0.5)
            else:
                depth_mean_even = numpy.nan
                depth_mean_even_std = numpy.nan

            in_transit_count, after_transit_count, before_transit_count = count_stats(
                transitleastsquares.t, transitleastsquares.y, transit_times, transit_duration_in_days
            )

            # Odd even mismatch in standard deviations
            odd_even_difference = abs(depth_mean_odd - depth_mean_even)
            odd_even_std_sum = depth_mean_odd_std + depth_mean_even_std
            odd_even_mismatch = odd_even_difference / odd_even_std_sum

            transit_count = len(transit_times)
            empty_transit_count = numpy.count_nonzero(per_transit_count == 0)
            distinct_transit_count = transit_count - empty_transit_count

            duration = transit_duration_in_days

            if empty_transit_count / transit_count >= 0.33:
                text = (
                    str(empty_transit_count)
                    + " of "
                    + str(transit_count)
                    + " transits without data. The true period may be twice the given period."
                )
                warnings.warn(text)

            rp_rs = rp_rs_from_depth(depth=1 - depth, law=transitleastsquares.limb_dark, params=transitleastsquares.u)
        return transitleastsquaresresults(
            SDE,
            SDE_raw,
            chi2_min,
            chi2red_min,
            period,
            period_uncertainty(test_statistic_periods, power),
            T0,
            duration,
            depth,
            (depth_mean, depth_mean_std),
            (depth_mean_even, depth_mean_even_std),
            (depth_mean_odd, depth_mean_odd_std),
            transit_depths,
            transit_depths_uncertainties,
            rp_rs,
            snr,
            snr_per_transit,
            snr_pink_per_transit,
            odd_even_mismatch,
            transit_times,
            per_transit_count,
            transit_count,
            distinct_transit_count,
            empty_transit_count,
            FAP(SDE),
            in_transit_count,
            after_transit_count,
            before_transit_count,
            test_statistic_periods,
            power,
            power_raw,
            SR,
            chi2,
            chi2red,
            model_lightcurve_time,
            model_lightcurve_model,
            model_folded_phase,
            folded_y,
            folded_dy,
            folded_phase,
            model_folded_model,
        )



import setuptools
from setuptools import setup
from os import path

# Pull TLS version from single source of truth file
try:  # Python 2
    execfile(path.join("foldedleastsquares", 'version.py'))
except:  # Python 3
    exec(open(path.join("foldedleastsquares", 'version.py')).read())


# If Python3: Add "README.md" to setup. 
# Useful for PyPI (pip install transitleastsquares). Irrelevant for users using Python2
try:
    
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
except:
    long_description = ' '

    
setup(name='foldedleastsquares',
    version=TLS_VERSIONING,
    description='An optimized transit-fitting algorithm to search for periodic features in light curves',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/hippke/tls',
    author='Martín Dévora Pajares',
    author_email='martin.devora.pajares@gmail.com',
    license='MIT',
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={'': ['*.csv', '*.cfg']},
    entry_points = {'console_scripts': ['transitleastsquares=transitleastsquares.command_line:main'],},
    install_requires=[
        'astropy<3;python_version<"3"',  # astropy 3 doesn't install in Python 2, but is req for astroquery
        'astroquery>=0.3.9',  # earlier has bug for "from astroquery.mast import Catalogs"
        'numpy',
        'numba',
        'tqdm',
        'batman-package',
        'argparse',
        'configparser'
        ]
)



from __future__ import division, print_function
import os
import numpy
import batman
from foldedleastsquares import transitleastsquares


if __name__ == "__main__":
    print("Starting test: synthetic...", end="")
    
    numpy.random.seed(seed=0)  # reproducibility
    # Create test data
    start = 48
    days = 365.25 * 3
    samples_per_day = 12  # 48
    samples = int(days * samples_per_day)  # 48
    t = numpy.linspace(start, start + days, samples)

    # Use batman to create transits
    ma = batman.TransitParams()
    ma.t0 = (
        start + 20
    )  # time of inferior conjunction; first transit is X days after start
    ma.per = 365.25  # orbital period
    ma.rp = 6371 / 696342  # 6371 planet radius (in units of stellar radii)
    ma.a = 217  # semi-major axis (in units of stellar radii)
    ma.inc = 90  # orbital inclination (in degrees)
    ma.ecc = 0  # eccentricity
    ma.w = 90  # longitude of periastron (in degrees)
    ma.u = [0.5]  # limb darkening coefficients
    ma.limb_dark = "linear"  # limb darkening model
    m = batman.TransitModel(ma, t)  # initializes model
    original_flux = m.light_curve(ma)  # calculates light curve

    # Create noise and merge with flux
    ppm = 5
    stdev = 10 ** -6 * ppm
    noise = numpy.random.normal(0, stdev, int(samples))
    y = original_flux + noise
    y[1] = numpy.nan
    model = transitleastsquares(t, y)
    results = model.power(
        period_min=360,
        period_max=370,
        transit_depth_min=10 * 10 ** -6,
        oversampling_factor=5,
        duration_grid_step=1.02,
    )

    numpy.testing.assert_almost_equal(results.chi2_min, 8831.654060613922, decimal=5)
    numpy.testing.assert_almost_equal(
        results.chi2red_min, 0.6719152511118321, decimal=5
    )

    numpy.testing.assert_almost_equal(
        results.period_uncertainty, 0.216212529678387, decimal=5
    )
    numpy.testing.assert_equal(results.per_transit_count[0], 7)
    numpy.testing.assert_equal(len(results.transit_times), 3)
    numpy.testing.assert_almost_equal(results.period, 365.2582192473641, decimal=5)
    numpy.testing.assert_almost_equal(
        results.transit_times[0], 68.00349264912924, decimal=5
    )
    """
    numpy.testing.assert_almost_equal(results.depth, 0.999897160189092, decimal=5)
    numpy.testing.assert_almost_equal(results.duration, 0.5908251624976649, decimal=5)
    numpy.testing.assert_almost_equal(
        min(results.chi2red), 0.6719167401148216, decimal=5
    )
    numpy.testing.assert_almost_equal(results.SDE, 5.691301613227594, decimal=5)
    numpy.testing.assert_almost_equal(
        results.odd_even_mismatch, 0.29083256866622437, decimal=5
    )
    numpy.testing.assert_almost_equal(results.rp_rs, 0.009119851811944274, decimal=5)

    # Full light curve model
    numpy.testing.assert_almost_equal(
        max(results.model_lightcurve_time), 1143.7472155961277, decimal=5
    )
    numpy.testing.assert_almost_equal(
        min(results.model_lightcurve_time), 48.0059010663453, decimal=5
    )
    numpy.testing.assert_almost_equal(
        numpy.mean(results.model_lightcurve_time), 595.877471821318, decimal=5
    )
    numpy.testing.assert_almost_equal(
        numpy.sum(results.model_lightcurve_time), 39166430.34534341, decimal=5
    )

    numpy.testing.assert_almost_equal(max(results.model_lightcurve_model), 1, decimal=5)
    numpy.testing.assert_almost_equal(
        min(results.model_lightcurve_model), 0.999897160189092, decimal=5
    )
    numpy.testing.assert_almost_equal(
        numpy.mean(results.model_lightcurve_model), 0.9999998641488729, decimal=5
    )
    numpy.testing.assert_almost_equal(
        numpy.sum(results.model_lightcurve_model), 65728.99107064126, decimal=5
    )

    transit_times_expected = [68.003492, 433.261711, 798.519931]
    numpy.testing.assert_almost_equal(
        results.transit_times, transit_times_expected, decimal=5
    )
    numpy.testing.assert_almost_equal(results.duration, 0.590825, decimal=5)

    numpy.testing.assert_almost_equal(
        max(results.model_folded_phase), 1.0000380285975052, decimal=5
    )
    numpy.testing.assert_almost_equal(
        min(results.model_folded_phase), 3.8028597505324e-05, decimal=5
    )
    numpy.testing.assert_almost_equal(
        numpy.mean(results.model_folded_phase), 0.5000380285975052, decimal=5
    )
    numpy.testing.assert_almost_equal(
        numpy.sum(results.model_folded_phase), 6574.499999999999, decimal=5
    )

    numpy.testing.assert_almost_equal(max(results.model_folded_model), 1, decimal=5)
    numpy.testing.assert_almost_equal(
        min(results.model_folded_model), 0.999897160189092, decimal=5
    )
    numpy.testing.assert_almost_equal(
        numpy.mean(results.model_folded_model), 0.9999998679702978, decimal=5
    )
    numpy.testing.assert_almost_equal(
        numpy.sum(results.model_folded_model), 13147.998264073476, decimal=5
    )

    numpy.testing.assert_almost_equal(
        max(results.folded_phase), 0.9999608485845858, decimal=5
    )
    numpy.testing.assert_almost_equal(
        min(results.folded_phase), 1.44015016259047e-05, decimal=5
    )
    numpy.testing.assert_almost_equal(
        numpy.mean(results.folded_phase), 0.500000089528271, decimal=5
    )
    numpy.testing.assert_almost_equal(
        numpy.sum(results.folded_phase), 6574.001177117707, decimal=5
    )

    numpy.testing.assert_almost_equal(
        max(results.folded_y), 1.000019008301075, decimal=5
    )
    numpy.testing.assert_almost_equal(
        min(results.folded_y), 0.9998860842491378, decimal=5
    )
    numpy.testing.assert_almost_equal(
        numpy.mean(results.folded_y), 0.9999997920032417, decimal=5
    )
    numpy.testing.assert_almost_equal(
        numpy.sum(results.folded_y), 13147.997265, decimal=5
    )

    numpy.testing.assert_almost_equal(
        results.depth_mean_even, (0.999915, 6.785539e-06), decimal=5
    )
    numpy.testing.assert_almost_equal(
        results.depth_mean_odd, (0.999920, 1.209993e-05), decimal=5
    )
    numpy.testing.assert_almost_equal(
        results.depth_mean, (0.999917, 6.086923e-06), decimal=5
    )

    numpy.testing.assert_almost_equal(
        results.transit_depths, [0.99991085, 0.99992095, 0.99992007], decimal=5
    )
    numpy.testing.assert_almost_equal(
        results.transit_depths_uncertainties,
        [4.19177855e-06, 1.20999330e-05, 1.26699399e-05],
        decimal=5,
    )
    numpy.testing.assert_almost_equal(
        results.odd_even_mismatch, 0.29083256866622437, decimal=5
    )
    numpy.testing.assert_almost_equal(
        results.per_transit_count, [7.0, 7.0, 7.0], decimal=5
    )
    numpy.testing.assert_almost_equal(results.transit_count, 3, decimal=5)
    numpy.testing.assert_almost_equal(results.distinct_transit_count, 3, decimal=5)
    numpy.testing.assert_almost_equal(results.empty_transit_count, 0, decimal=5)
    numpy.testing.assert_almost_equal(
        results.snr_per_transit, [38.92162, 34.51048, 34.89514], decimal=5
    )
    numpy.testing.assert_almost_equal(results.snr, 62.542764907612785, decimal=5)
    numpy.testing.assert_almost_equal(
        results.snr_pink_per_transit, [52.24377, 46.32278, 46.8391], decimal=5
    )
    """
    print("passed")



from __future__ import division, print_function
import numpy
from foldedleastsquares.template_generator.default_transit_template_generator import DefaultTransitTemplateGenerator

if __name__ == "__main__":
    print("Starting test: duration_grid...", end="")
    default_transit_template_generator = DefaultTransitTemplateGenerator()
    periods = default_transit_template_generator.period_grid(
        R_star=1,  # R_sun
        M_star=1,  # M_sun
        time_span=20,  # days
        period_min=0,
        period_max=999,
        oversampling_factor=3,
    )
    durations = DefaultTransitTemplateGenerator().duration_grid(periods, log_step=1.05, shortest=2)
    numpy.testing.assert_almost_equal(max(durations), 0.12)
    numpy.testing.assert_almost_equal(min(durations), 0.004562690993268325)
    numpy.testing.assert_equal(len(durations), 69)
    print("passed")
